# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the class command group which includes
commands for get, enumerate, associators, references, find, etc. of the objects
CIMClass on a WBEM server.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import click

from pywbem import Error, CIMClassName, CIMError, ModelError, \
    CIM_ERR_NOT_FOUND, CIMClass

from .pywbemcli import cli
from ._common import display_cim_objects, filter_namelist, \
    resolve_propertylist, CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT, \
    output_format_is_table, format_table, process_invokemethod, \
    raise_pywbem_error_exception, warning_msg, validate_output_format
from ._common_options import add_options, propertylist_option, \
    names_only_option, include_classorigin_class_option, namespace_option,  \
    summary_option, multiple_namespaces_option, association_filter_option, \
    indication_filter_option, experimental_filter_option, help_option
from ._displaytree import display_class_tree
from ._click_extensions import PywbemcliGroup, PywbemcliCommand


#
#   Common option definitions for class group
#

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

no_qualifiers_class_option = [              # pylint: disable=invalid-name
    click.option('--nq', '--no-qualifiers', 'no_qualifiers', is_flag=True,
                 default=True,
                 help=u'Do not include qualifiers in the returned class(es). '
                      'Default: Include qualifiers.')]

deep_inheritance_class_option = [              # pylint: disable=invalid-name
    click.option('--di', '--deep-inheritance', 'deep_inheritance', is_flag=True,
                 default=False,
                 help=u'Include the complete subclass hierarchy of the '
                      'requested classes in the result set. '
                      'Default: Do not include subclasses.')]

local_only_class_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 default=False,
                 help=u'Do not include superclass properties and methods in '
                      u'the returned class(es). '
                      u'Default: Include superclass properties and methods.')]


##########################################################################
#
#   Click command group and command definitions
#   These decorated functions implement the commands, arguments, and
#   options for the top-level class command group
#
###########################################################################

@cli.group('class', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def class_group():
    """
    Command group for CIM classes.

    This command group defines commands to inspect classes, invoke
    methods on classes, delete classes.

    Creation and modification of classes is not currently supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'class' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@class_group.command('enumerate', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@add_options(deep_inheritance_class_option)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(association_filter_option)
@add_options(indication_filter_option)
@add_options(experimental_filter_option)
@add_options(help_option)
@click.pass_obj
def class_enumerate(context, classname, **options):
    """
    List top classes or subclasses of a class in a namespace.

    Enumerate CIM classes starting either at the top of the class hierarchy
    in the specified CIM namespace (--namespace option), or at the specified
    class (CLASSNAME argument) in the specified namespace. If no namespace was
    specified, the default namespace of the connection is used.

    The --local-only, --include-classorigin, and --no-qualifiers options
    determine which parts are included in each retrieved class.

    The --deep-inheritance option defines whether or not the complete subclass
    hierarchy of the classes is retrieved.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class enumerate -n interop

      pywbemcli -n myconn class enumerate CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_enumerate(context, classname,
                                                    options))


@class_group.command('get', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@add_options(local_only_class_option)
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_get(context, classname, **options):
    """
    Get a class.

    Get a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    The --local-only, --include-classorigin, --no-qualifiers, and
    --propertylist options determine which parts are included in each retrieved
    class.

    In the output, the class will be formatted as defined by the
    --output-format general option. Table formats are replaced with MOF
    format.

    Example:

      pywbemcli -n myconn class get CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_get(context, classname, options))


@class_group.command('delete', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.option('-f', '--force', is_flag=True, default=False,
              help=u'Delete any instances of the class as well. '
                   'Some servers may still reject the class deletion. '
                   'Default: Reject command if the class has any instances.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_delete(context, classname, **options):
    """
    Delete a class.

    Delete a CIM class (CLASSNAME argument) in a CIM namespace (--namespace
    option). If no namespace was specified, the default namespace of the
    connection is used.

    If the class has subclasses, the command is rejected.

    If the class has instances, the command is rejected, unless the --force
    option was specified, in which case the instances are also deleted.

    WARNING: Deleting classes can cause damage to the server: It can impact
    instance providers and other components in the server. Use this
    command with caution.

    Many WBEM servers may not allow this operation or may severely limit the
    conditions under which a class can be deleted from the server.

    Example:

      pywbemcli -n myconn class delete CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_delete(context, classname, options))


@class_group.command('invokemethod', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True,)
@click.argument('methodname', type=str, metavar='METHODNAME', required=True)
@click.option('-p', '--parameter', type=str, metavar='PARAMETERNAME=VALUE',
              required=False, multiple=True,
              help=u'Specify a method input parameter with its value. '
                   'May be specified multiple times. '
                   'Default: No input parameters.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_invokemethod(context, classname, methodname, **options):
    """
    Invoke a method on a class.

    Invoke a static CIM method (METHODNAME argument) on a CIM class (CLASSNAME
    argument) in a CIM namespace (--namespace option), and display the method
    return value and output parameters. If no namespace was specified, the
    default namespace of the connection is used.

    The method input parameters are specified using the --parameter option,
    which may be specified multiple times.

    Pywbemcli retrieves the class definition from the server in order to
    verify that the specified input parameters are consistent with the
    parameter characteristics in the method definition.

    Use the 'instance invokemethod' command to invoke CIM methods on CIM
    instances.

    Example:

      pywbemcli -n myconn class invokemethod CIM_Foo methodx -p p1=9 -p p2=Fred
    """
    context.execute_cmd(lambda: cmd_class_invokemethod(context,
                                                       classname,
                                                       methodname,
                                                       options))


@class_group.command('references', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def class_references(context, classname, **options):
    """
    List the classes referencing a class.

    List the CIM (association) classes that reference the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role and
    --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class references CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_references(context, classname,
                                                     options))


@class_group.command('associators', cls=PywbemcliCommand,
                     options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help=u'Filter the result set by association class name. '
                   'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by far end role name.')
@add_options(no_qualifiers_class_option)
@add_options(include_classorigin_class_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def class_associators(context, classname, **options):
    """
    List the classes associated with a class.

    List the CIM classes that are associated with the specified class
    (CLASSNAME argument) in the specified CIM namespace
    (--namespace option). If no namespace was specified, the default namespace
    of the connection is used.

    The classes to be retrieved can be filtered by the --role, --result-role,
    --assoc-class, and --result-class options.

    The --include-classorigin, --no-qualifiers, and --propertylist options
    determine which parts are included in each retrieved class.

    The --names-only option can be used to show only the class paths.

    In the output, the classes and class paths will be formatted as defined
    by the --output-format general option. Table formats on classes will be
    replaced with MOF format.

    Examples:

      pywbemcli -n myconn class associators CIM_Foo -n interop
    """
    context.execute_cmd(lambda: cmd_class_associators(context, classname,
                                                      options))


@class_group.command('find', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname-glob', type=str, metavar='CLASSNAME-GLOB',
                required=True)
@add_options(multiple_namespaces_option)
@click.option('-s', '--sort', is_flag=True, required=False,
              help=u'Sort by namespace. Default is to sort by classname')
@add_options(association_filter_option)
@add_options(indication_filter_option)
@add_options(experimental_filter_option)
@add_options(help_option)
@click.pass_obj
def class_find(context, classname_glob, **options):
    """
    List the classes with matching class names on the server.

    Find the CIM classes whose class name matches the specified wildcard
    expression (CLASSNAME-GLOB argument) in all CIM namespaces of the
    WBEM server, or in the specified namespace (--namespace option).

    The CLASSNAME-GLOB argument is a wildcard expression that is matched on
    class names case insensitively.
    The special characters from Unix file name wildcarding are supported
    ('*' to match zero or more characters, '?' to match a single character,
    and '[]' to match character ranges). To avoid shell expansion of wildcard
    characters, the CLASSNAME-GLOB argument should be put in quotes.

    For example, "pywbem_*" returns classes whose name begins with "PyWBEM_",
    "pywbem_", etc. "*system*" returns classes whose names include the case
    insensitive string "system".

    In the output, the classes will be formatted as defined by the
    --output-format general option if it specifies table output. Otherwise
    the classes will be in the form "NAMESPACE:CLASSNAME".

    Examples:

      pywbemcli -n myconn class find "CIM_*System*" -n interop

      pywbemcli -n myconn class find *Foo*
    """
    context.execute_cmd(lambda: cmd_class_find(context, classname_glob,
                                               options))


@class_group.command('tree', cls=PywbemcliCommand, options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=False)
@click.option('-s', '--superclasses', is_flag=True, default=False,
              help=u'Show the superclass hierarchy. '
                   'Default: Show the subclass hierarchy.')
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def class_tree(context, classname, **options):
    """
    Show the subclass or superclass hierarchy for a class.

    List the subclass or superclass hierarchy of a CIM class (CLASSNAME
    argument) or CIM namespace (--namespace option):

    - If CLASSNAME is omitted, the complete class hierarchy of the specified
      namespace is retrieved.

    - If CLASSNAME is specified but not --superclasses, the class and its
      subclass hierarchy in the specified namespace are retrieved.

    - If CLASSNAME and --superclasses are specified, the class and its
      superclass ancestry up to the top-level class in the specified namespace
      are retrieved.

    If no namespace was specified, the default namespace of the connection is
    used.

    In the output, the classes will formatted as a ASCII graphical tree; the
    --output-format general option is ignored.

    Examples:

      pywbemcli -n myconn class tree -n interop

      pywbemcli -n myconn class tree CIM_Foo -n interop

      pywbemcli -n myconn class tree CIM_Foo -s -n interop
    """
    context.execute_cmd(lambda: cmd_class_tree(context, classname, options))


####################################################################
#
#  Common functions for cmd_class processing
#  This includes functions used by the command action functions
#  in this fmodule and possibly other modules
#
####################################################################


def _build_qualifier_filters(options):
    """
    Build a dictionary defining the qualifier filters to be processes from
    their definitons in the Click options dictionary. There is an entry
    in the dictionary for each qualifier filter where the key is the
    association name and the value is True or False depending in the
    value of the option ('x' or 'no-x' )
    """
    qualifier_filters = {}
    if options['association'] is not None:
        # show_assoc = options['association']
        qualifier_filters['Association'] = options['association']
    if options['indication'] is not None:
        qualifier_filters['Indication'] = options['indication']
    if options['experimental'] is not None:
        qualifier_filters['Experimental'] = options['experimental']
    return qualifier_filters


def _filter_classes_for_qualifiers(qualifier_filters, results, names_only, iq):
    """
    Filter the results list for the qualifiers defined by filter
    qualifier: a dictionary with qualifier name as key and Boolean defining
    whether to display or not display if it exists.

    This method only works for boolean qualifiers
    """

    filtered_results = []
    if results:
        assert isinstance(results[0], CIMClass)
    for cls in results:
        assert isinstance(cls, CIMClass)
        show_this_class = True
        for qname, show_if_true in qualifier_filters.items():
            if qname in cls.qualifiers:
                qvalue = cls.qualifiers[qname].value
                show_this = (qvalue == show_if_true)
            else:
                show_this = not show_if_true
            if not show_this:
                show_this_class = False
                break
        if show_this_class:
            # If returning instances, honor the names_only option
            if not names_only:
                if not iq:
                    cls.qualifiers = []
                    for p in cls.properties.values():
                        p.qualifiers = []
                    for m in cls.methods.values():
                        m.qualifiers = []
                        for p in m.parameters.values():
                            p.qualifiers = []
            filtered_results.append(cls)
    if names_only:
        filtered_results = [cls.classname for cls in filtered_results]
    return filtered_results


def enumerate_classes_filtered(context, classname, options):
    """
    Execute EnumerateClasses or EnumerateClassNames in a single namespace
    defined in options['namespace'] and return results.

    If any of the class qualifier filters are defined in the options parameter,
    enumerate the classes, filter the result for those parameters, and return
    only class names if --names-only set.

    This function may be executed by multiple command action functions with
    varying options in the options. Each option must be tested to validate
    that it exists in the options dictionary

    Parameters:

      context:  Click context

      classname:
        Optional classname for the enumerate.

      options:Click options dictionary
        Options that form basis for this Enumerate and filter processing.

    Returns:
        List of classes or classnames that satisfy the criteria

    Raises:
        pywbem Error exceptions generated by EnumerateClassNames and
        enumerateClasses
    """
    qualifier_filters = _build_qualifier_filters(options)

    names_only = options.get('names_only', False)

    iq = options.get('no_qualifiers', True)

    # Force IncludeQualifier true if results are to be filtered since
    # the filter requires that qualifiers exist.
    request_iq = True if qualifier_filters else iq

    local_only = options.get('local_only', False)
    deep_inheritance = options.get('deep_inheritance', True)
    include_classorigin = options.get('include_classorigin', True)

    if names_only and not qualifier_filters:
        results = context.conn.EnumerateClassNames(
            ClassName=classname,
            namespace=options['namespace'],
            DeepInheritance=deep_inheritance)
    else:
        results = context.conn.EnumerateClasses(
            ClassName=classname,
            namespace=options['namespace'],
            LocalOnly=local_only,
            DeepInheritance=deep_inheritance,
            IncludeQualifiers=request_iq,
            IncludeClassOrigin=include_classorigin)
        if qualifier_filters:
            results = _filter_classes_for_qualifiers(
                qualifier_filters, results,
                names_only, iq)
    return results


def get_format_group(context, options):
    """
    Define the format groups allowed based on the options. This is particular
    to the class commands, largely because we do not have a table format
    for enumerate, get, associators, or references unless options such as
    summary, or names_only are set.
    """

    # Summary always output as TABLE
    if 'summary' in options and options['summary']:

        # This accounts for the fact that the results of a summary can be
        # either table or simply a string output
        if context.output_format and \
                output_format_is_table(context.output_format):
            return ['TABLE']

        # Temporary hack. We need another format group, i.e. txt or str
        # That displays in non-structured manner. Or drop this output
        # completely.
        return ['CIM']

    # Names_only may be output as Table or as CIM Object.
    if 'names_only' in options and options['names_only']:
        return ['CIM', 'TABLE']

    # otherwise only CIM allowed today.
    return ['CIM']


#####################################################################
#
#  Command functions for each of the commands in the class group
#
#####################################################################


def cmd_class_get(context, classname, options):
    """
    Get the class defined by the argument.

    Gets the class defined by CLASSNAME from the WBEM server and displays
    the class. If the class cannot be found, the server returns a CIMError
    exception.
    """
    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        result_class = context.conn.GetClass(
            classname,
            namespace=options['namespace'],
            LocalOnly=options['local_only'],
            IncludeQualifiers=options['no_qualifiers'],
            IncludeClassOrigin=options['include_classorigin'],
            PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, result_class,
                            output_format=output_format)
    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_class_invokemethod(context, classname, methodname, options):
    """
    Create an instance and submit to a WBEM server
    """
    try:
        process_invokemethod(context, classname, methodname, options)
    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_class_enumerate(context, classname, options):
    """
        Enumerate the classes returning a list of classes from the WBEM server.
        That match the qualifier filter options
    """
    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        results = enumerate_classes_filtered(context, classname, options)

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_class_references(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        if options['names_only']:
            results = context.conn.ReferenceNames(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'])
        else:
            results = context.conn.References(
                classname,
                ResultClass=options['result_class'],
                Role=options['role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_class_associators(context, classname, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    format_group = get_format_group(context, options)
    output_format = validate_output_format(context.output_format, format_group)

    try:
        if options['names_only']:
            results = context.conn.AssociatorNames(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'])
        else:
            results = context.conn.Associators(
                classname,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                IncludeQualifiers=options['no_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_format,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)


def get_namespaces(context, namespaces):
    """
    Returns either the namespaces provided or if that is None, the set of
    namespaces that are defined in the wbem server as a list

    Raises:
        CIMError if status code not CIM_ERR_NOT_FOUND
    """
    ns_names = []

    # Return the provided namespace(s)
    if namespaces:
        return namespaces

    # Otherwise get all namespaces from server
    try:
        ns_names = context.wbem_server.namespaces
        ns_names.sort()
        return ns_names

    except ModelError:
        return [context.conn.default_namespace]
    except CIMError as ce:
        # allow processing to continue if no interop namespace
        if ce.status_code == CIM_ERR_NOT_FOUND:
            warning_msg('{}. Using default_namespace {}.'
                        .format(ce, context.conn.default_namespace))
            ns_names = [context.conn.default_namespace]
        return ns_names
    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_class_find(context, classname_glob, options):
    """
    Execute the command for enumerate classes, filter the results based on the
    option and display the result. The result is a list of classes/namespaces
    """

    output_format = validate_output_format(context.output_format, 'TABLE')

    context.spinner_stop()
    namespaces = get_namespaces(context, options['namespace'])

    try:
        names_dict = {}
        if namespaces:
            for namespace in namespaces:
                # Set cmd options that are required for this command to get
                # information on classes in server.
                # 1. Always use deep_inheritance
                # 2. Set namespace to each namespace in loop
                options['deep_inheritance'] = True
                options['namespace'] = namespace
                options['names_only'] = True

                classnames = enumerate_classes_filtered(context, None, options)
                names_dict[namespace] = filter_namelist(classname_glob,
                                                        classnames)

        # build rows of namespace, classname for each namespace, sort if
        # necessary,  and add to common rows
        row = 0 if options['sort'] else 1
        rows = []
        for ns_name in names_dict:
            ns_rows = [[ns_name, name] for name in names_dict[ns_name]]
            # sort by classname if sort option defined, else by namespace
            ns_rows.sort(key=lambda x: x[row])
            rows.extend(ns_rows)

        context.spinner_stop()
        if output_format_is_table(context.output_format):
            headers = ['Namespace', 'Classname']
            click.echo(
                format_table(rows, headers,
                             title='Find class {}'.format(classname_glob),
                             table_format=output_format))
        else:
            # Display function to display classnames returned with
            # their namespaces in the form <namespace>:<classname>
            context.spinner_stop()
            for row in rows:
                click.echo('  {}:{}'.format(row[0], row[1]))

    except Error as er:
        raise_pywbem_error_exception(er)


def get_class_hierarchy(conn, classname, namespace, superclasses=None):
    """
    Get the class hierarchy from the server, either the superclasses
    associated or subclasseswith classname .  If getting subclasses
    the classname parameter may be None which requests that the complete
    class hiearchy be retrieved.

      Parameters:
        conn (:class:`~pywbem.WBEMConnection` or subclass):
            Current connection to a WBEM server.

        classname(:term:`string):
            classname if the tree is to be initiated from
            within the class hiearchy. May be None

        namespace(:term:`string):
          Namespace to use to acquire classes from WBEM server

        superclasses (:class:`py:bool`):
            If `True` display the superclasses of classname. If not True
            display the subclasses. The default is None (display subclasses).

      Returns:
        tuple of  classes, classnames where
          classes (list of :class:`~pywbem.CIMClass`): that are either the
          superclasses or the subclasses of classname.

      Raises:
        CIM_Error:
    """

    try:
        if superclasses:
            # classname must exist for superclass tree
            assert classname is not None

            # get the superclasses into a list
            klass = conn.GetClass(classname, namespace=namespace)
            classes = []
            classes.append(klass)
            while klass.superclass:
                klass = conn.GetClass(klass.superclass, namespace=namespace)
                classes.append(klass)
        else:
            # get the subclass hierarchy either complete or starting at the
            # optional CLASSNAME. Gets minimum data from server to define
            # class, superclass data
            classes = conn.EnumerateClasses(ClassName=classname,
                                            namespace=namespace,
                                            LocalOnly=True,
                                            IncludeQualifiers=False,
                                            DeepInheritance=True)

    except Error as er:
        raise_pywbem_error_exception(er)

    return classes


def cmd_class_tree(context, classname, options):
    """
    Execute the command to display graphical tree of either superclasses or
    subclasses of classname. If classname is None, display tree starting from
    the class hierarchy root.
    Tree is displayed on the console as a left-justified tree using the
    asciitree library.
    The --superclasses option determines if the superclass tree or the
    subclass tree is displayed.
    """
    superclasses = options['superclasses']

    # Classname must exist as starting point for superclass tree.
    if superclasses and classname is None:
        raise click.ClickException('CLASSNAME argument required for '
                                   '--superclasses option')

    classes = get_class_hierarchy(context.conn, classname, options['namespace'],
                                  superclasses)

    # If showing superclasses, set classname to None for the display
    # to add the 'root' as top class.
    if superclasses:
        classname = None

    # Display the list of classes as a tree. The classname is the top
    # of the tree.
    context.spinner_stop()
    display_class_tree(classes, classname)


def cmd_class_delete(context, classname, options):
    """Delete a class from the WBEM server repository"""
    if options['namespace']:
        classname = CIMClassName(classname, namespace=options['namespace'])

    try:
        instnames = context.conn.EnumerateInstanceNames(classname)
        subclassnames = context.conn.EnumerateClassNames(ClassName=classname,
                                                         DeepInheritance=True)
    except Error as er:
        raise_pywbem_error_exception(er)

    if subclassnames:
        raise click.ClickException('Delete rejected; subclasses exist')

    if not options['force']:
        if instnames:
            raise click.ClickException('Delete rejected; instances exist')
    else:
        for instname in instnames:
            context.conn.DeleteInstance(instname)

    instnames = context.conn.EnumerateInstanceNames(classname)
    if instnames:
        raise click.ClickException('Delete rejected; instance delete failed')

    try:
        context.conn.DeleteClass(classname)
        if context.verbose:
            context.spinner_stop()
            click.echo('Deleted class {}.'.format(classname))
    except Error as er:
        raise_pywbem_error_exception(er)



# (C) Copyright 2020 IBM Corp.
# (C) Copyright 2020 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Initialization of the pywbemtools package.
"""

import sys

# Keep these Python versions in sync with setup.py
_python_m = sys.version_info[0]  # pylint: disable=invalid-name
_python_n = sys.version_info[1]  # pylint: disable=invalid-name
if _python_m == 2 and _python_n < 7:
    raise RuntimeError('On Python 2, pywbemtools requires Python 2.7 or higher')
if _python_m == 3 and _python_n < 4:
    raise RuntimeError('On Python 3, pywbemtools requires Python 3.4 or higher')



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click command definition for the server command group which includes
cmds for inspection and management of the objects defined by the pywbem
server class including namespaces, WBEMServer information, and profile
information.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import click

from pywbem import ValueMapping, Error

from .pywbemcli import cli
from ._common import format_table, raise_pywbem_error_exception, \
    validate_output_format, display_text, output_format_is_table, \
    CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT, DEFAULT_TABLE_FORMAT
from ._common_options import add_options, help_option
from ._click_extensions import PywbemcliGroup, PywbemcliCommand

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue


@cli.group('server', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def server_group():
    """
    Command group for WBEM servers.

    This command group defines commands to inspect and manage core components
    of a WBEM server including server attributes, namespaces, the Interop
    namespace, management profiles, and access to profile central instances.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'server' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@server_group.command('namespaces', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_namespaces(context):
    """
    List the namespaces of the server.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_server_namespaces(context))


@server_group.command('interop', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_interop(context):
    """
    Get the Interop namespace of the server.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_server_interop(context))


@server_group.command('brand', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_brand(context):
    """
    Get the brand of the server.

    Brand information is defined by the server implementor and may or may
    not be available. Pywbem attempts to collect the brand information from
    multiple sources.
    """
    # pylint: disable=too-many-function-args
    context.execute_cmd(lambda: cmd_server_brand(context))


@server_group.command('info', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@add_options(help_option)
@click.pass_obj
def server_info(context):
    """
    Get information about the server.

    The information includes CIM namespaces and server brand.
    """
    context.execute_cmd(lambda: cmd_server_info(context))


@server_group.command('profiles', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@click.option('-o', '--organization', type=str, metavar='ORG-NAME',
              required=False,
              help='Filter by the defined organization. (ex. -o DMTF')
@click.option('-p', '--profile', type=str, metavar='PROFILE-NAME',
              required=False,
              help='Filter by the profile name. (ex. -p Array')
@add_options(help_option)
@click.pass_obj
def server_profiles(context, **options):
    """
    List management profiles advertized by the server.

    Retrieve the CIM instances representing the WBEM management profiles
    advertized by the WBEM server, and display information about each profile.
    WBEM management profiles are defined by DMTF and SNIA and define the
    management functionality that is available.

    The retrieved profiles can be filtered using the --organization and
    --profile options.

    The output is formatted as a table showing the organization, name, and
    version for each profile. The --output-format option is ignored unless it
    specifies a table format.
    """
    context.execute_cmd(lambda: cmd_server_profiles(context, options))


@server_group.command('centralinsts', cls=PywbemcliCommand,
                      options_metavar=CMD_OPTS_TXT)
@click.option('-o', '--organization', type=str, metavar='ORG-NAME',
              required=False,
              help='Filter by the defined organization. (ex. -o DMTF')
@click.option('-p', '--profile', type=str, metavar='PROFILE-NAME',
              required=False,
              help='Filter by the profile name. (ex. -p Array')
@click.option('--cc', '--central-class', 'central_class', type=str,
              metavar='CLASSNAME', required=False,
              help='Optional. Required only if profiles supports only '
              'scopig methodology')
@click.option('--sc', '--scoping-class', 'scoping_class', type=str,
              metavar='CLASSNAME', required=False,
              help='Optional. Required only if profiles supports only '
              'scopig methodology')
@click.option('--sp', '--scoping-path', 'scoping_path', type=str,
              metavar='CLASSLIST', required=False, multiple=True,
              help='Optional. Required only if profiles supports only '
              'scopig methodology. Multiples allowed')
@click.option('--rd', '--reference-direction', 'reference_direction',
              type=click.Choice(['snia', 'dmtf']),
              default='dmtf',
              show_default=True,
              help='Navigation direction for association.')
@add_options(help_option)
@click.pass_obj
def centralinsts(context, **options):
    """
    List central instances of mgmt profiles on the server.

    Retrieve the CIM instances that are central instances of the specified
    WBEM management profiles, and display these instances. By default, all
    management profiles advertized on the server are used. The profiles
    can be filtered by using the --organization and --profile options.

    The central instances are determined using all methodologies defined
    in DSP1033 V1.1 in the order of GetCentralInstances, central class,
    and scoping class methodology.

    Profiles that only use the scoping class methodology require the
    specification of the --central-class, --scoping-class, and --scoping-path
    options because additional information is needed to perform the scoping
    class methodology.

    The retrieved central instances are displayed along with the organization,
    name, and version of the profile they belong to, formatted as a table.
    The --output-format general option is ignored unless it specifies a table
    format.
    """
    context.execute_cmd(lambda: cmd_server_centralinsts(context, options))


###############################################################
#         Server cmds
###############################################################


def cmd_server_namespaces(context):
    """
    Display namespaces in the current WBEM server
    """
    output_format = validate_output_format(
        context.output_format,
        ['TABLE', 'TEXT'], default_format=DEFAULT_TABLE_FORMAT)

    try:
        namespaces = context.wbem_server.namespaces
        namespaces.sort()
        context.spinner_stop()
        if output_format_is_table(output_format):
            # create list for each row
            rows = [[ns] for ns in namespaces]

            click.echo(format_table(rows, ['Namespace Name'],
                                    title='Server Namespaces:',
                                    table_format=output_format))
        else:
            display_text(", ".join(namespaces))

    except Error as er:
        raise click.ClickException('{}: {}'.format(er.__class__.__name__, er))


def cmd_server_interop(context):
    """
    Display interop namespace in the current WBEM server
    """

    output_format = validate_output_format(context.output_format, 'TEXT')

    try:
        interop_ns = context.wbem_server.interop_ns
        context.spinner_stop()

        display_text(interop_ns, output_format)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_server_brand(context):
    """
    Display product and version info of the current WBEM server
    """
    output_format = validate_output_format(context.output_format, 'TEXT')

    try:
        brand = context.wbem_server.brand
        context.spinner_stop()

        display_text(brand, output_format)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_server_info(context):
    """
    Display general overview of info from current WBEM server
    """
    output_format = validate_output_format(context.output_format, 'TABLE')

    try:
        # execute the namespaces to force contact with server before
        # turning off the spinner.
        server = context.wbem_server
        server.namespaces  # pylint: disable=pointless-statement
        context.spinner_stop()

        server = context.wbem_server

        rows = []
        headers = ['Brand', 'Version', 'Interop Namespace', 'Namespaces']
        if len(server.namespaces) > 3:
            namespaces = '\n'.join(server.namespaces)
        else:
            namespaces = ', '.join(server.namespaces)
        rows.append([server.brand, server.version,
                     server.interop_ns,
                     namespaces])
        click.echo(format_table(rows, headers,
                                title='Server General Information',
                                table_format=output_format))

    except Error as er:
        raise_pywbem_error_exception(er)


def get_profile_info(org_vm, inst):
    """
    Get the org, name, and version from the profile instance and
    return them as a tuple.
    """
    org = org_vm.tovalues(inst['RegisteredOrganization'])
    name = inst['RegisteredName']
    vers = inst['RegisteredVersion']
    return org, name, vers


def profile_row_sortkey(row):
    """
    Return a sort key for profiles that sorts by org name, profile name, and
    profile version (numeric).

    row is a tuple(org, name, version) where version is a version string such
    as "1.2.3" or "1.2.3a".
    """
    version_parts = []
    for v_str in row[2].split('.'):
        try:
            v_int = int(v_str)
        except ValueError:
            v_int = v_str
        version_parts.append(v_int)
    return row[0], row[1], tuple(version_parts)


def cmd_server_profiles(context, options):
    """
    Display general overview of info from current WBEM server
    """
    output_format = validate_output_format(context.output_format, 'TABLE')

    server = context.wbem_server
    try:
        found_server_profiles = server.get_selected_profiles(
            registered_org=options['organization'],
            registered_name=options['profile'])

        org_vm = ValueMapping.for_property(server,
                                           server.interop_ns,
                                           'CIM_RegisteredProfile',
                                           'RegisteredOrganization')
        rows = []
        for inst in found_server_profiles:
            row = get_profile_info(org_vm, inst)
            rows.append(row)
        rows.sort(key=profile_row_sortkey)
        headers = ['Organization', 'Registered Name', 'Version']

        click.echo(format_table(rows, headers,
                                title='Advertised management profiles:',
                                table_format=output_format))

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_server_centralinsts(context, options):
    """
    Display general information on the central instances of one or more
    profiles.
    """
    output_format = validate_output_format(context.output_format, ['CIM',
                                                                   'TABLE'])

    server = context.wbem_server
    try:
        found_server_profiles = server.get_selected_profiles(
            registered_org=options['organization'],
            registered_name=options['profile'])

        org_vm = ValueMapping.for_property(server,
                                           server.interop_ns,
                                           'CIM_RegisteredProfile',
                                           'RegisteredOrganization')
        rows = []
        for inst in found_server_profiles:
            pi = get_profile_info(org_vm, inst)
            row = [":".join(pi)]
            try:
                ci = server.get_central_instances(
                    inst.path,
                    central_class=options['central_class'],
                    scoping_class=options['scoping_class'],
                    scoping_path=options['scoping_path'],
                    reference_direction=options['reference_direction'])
                row.append("\n".join([str(p) for p in ci]))
            # mark current inst as failed and continue
            except Exception as ex:  # pylint: disable=broad-except
                click.echo('Exception: {} {}'.format(row, ex))
                row.append("Failed")
            rows.append(row)

        # sort by org
        rows.sort(key=lambda x: (x[0]))
        headers = ['Profile', 'Central Instance paths']

        click.echo(format_table(rows,
                                headers,
                                title='Advertised Central Instances:',
                                table_format=output_format))
    except Error as er:
        raise_pywbem_error_exception(er)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
pybemcli context object. This is the common object for click command calls for
pywbemcli context information.

It contains data that is set in the top level and used in command calls

This object is attached to the Click Context in pywbemcli.py
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import click
import click_spinner

from ._common import format_table


class ContextObj(object):  # pylint: disable=useless-object-inheritance
    """
        Manage the pywbemcli context that is carried within the Click
        context object in the obj parameter. This is the object that
        communicates between the cli commands and command groups. It contains
        the information that is common to the multiple click commands
    """

    spinner_envvar = 'PYWBEMCLI_SPINNER'

    # pylint: disable=unused-argument
    def __init__(self, pywbem_server, output_format, use_pull,
                 pull_max_cnt, timestats, log, verbose, pdb,
                 deprecation_warnings, connections_repo):

        self._pywbem_server = pywbem_server
        self._output_format = output_format
        self._use_pull = use_pull
        self._pull_max_cnt = pull_max_cnt
        self._timestats = timestats
        self._log = log
        self._verbose = verbose
        self._pdb = pdb
        self._deprecation_warnings = deprecation_warnings
        self._connections_repo = connections_repo

        self._spinner_enabled = None  # Deferred init in getter
        self._spinner_obj = click_spinner.Spinner()
        self._conn = None
        self._wbem_server = None

    def __repr__(self):
        return 'ContextObj(at {:08x}, pywbem_server={!r}, outputformat={}, ' \
               'use_pull={}, pull_max_cnt={}, timestats={}, verbose={}, ' \
               'spinner_enabled={}' \
               .format(id(self), self.pywbem_server, self.output_format,
                       self.use_pull, self.pull_max_cnt, self.timestats,
                       self.verbose, self.spinner_enabled)

    @property
    def output_format(self):
        """
        :term:`string`: String defining the output format requested.  This may
        be None meaning that the default format should be used or may be
        one of the values in the TABLE_FORMATS variable.
        """
        return self._output_format

    @property
    def timestats(self):
        """
        :term:`string`: Output format to be used.
        """
        return self._timestats

    @property
    def use_pull(self):
        """
        :term:`string`: Choice of whether pull, traditional or either type
        of operation is to be used for the instance enumerates, references,
        or associator commands.
        """
        return self._use_pull

    @property
    def pull_max_cnt(self):
        """
        :term:`string`: Maximum number of objects to be returne for pull op.
        """
        return self._pull_max_cnt

    @property
    def log(self):
        """
        :term:`string`: log definition from cmd line
        """
        return self._log

    @property
    def pdb(self):
        """
        bool: Indicates whether to break in the debugger.
        """
        return self._pdb

    @property
    def deprecation_warnings(self):
        """
        bool: Indicates whether to enable deprecation warnings.
        """
        return self._deprecation_warnings

    @property
    def connections_repo(self):
        """
        :class:`ConnectionRepository` instance defining at least the filename
        of the Connections repository. The connections_repo may not be
        loaded at this point.
        """
        return self._connections_repo

    @property
    def conn(self):
        """
        :class:`~pywbem.WBEMConnection` WBEMConnection to be used for requests.
        This property uses the wbem_server property to activate the
        conn and wbem_server. The connection is not activated unless
        a command requiring the webem server is executed.  This allows other
        commands like connection to execute without testing whether or not the
        WBEM server exists.
        """
        # The conn property is created in wbem_server and retained here.
        if self._conn:
            return self._conn

        self._conn = self.wbem_server.conn
        return self._conn

    @property
    def wbem_server(self):
        """
        :class:`~pywbem.WBEMServer` WBEMServer instance to be used for
        requests This is maintained in the pywbem_server object as
        _pywbem_server. This and/or the conn property are executed before
        any operation that is to contact the server.  This property enables any
        server characteristics. This is a passthrough for the wbem_server
        property maintained in self._pywbem_server
        """
        # If no server defined, do not try to connect. This allows
        # commands like help, connection new, list to execute without
        # a target server defined.
        if self._pywbem_server:  # pylint: disable=no-else-return
            # If wbem_server not initialized, initialize it.
            if self._pywbem_server.wbem_server is None:
                # get the password if it is required.  This may involve a
                # prompt.
                # TODO there are two creates and also, since all the inputs
                # are self, why do that here
                self._pywbem_server.get_password(self)
                self._pywbem_server.create_connection(
                    log=self.log,
                    use_pull=self.use_pull,
                    timestats=self.timestats,
                    verbose=self.verbose)
                if self._conn and self.timestats:  # Enable stats gathering
                    self.conn.statistics.enable()
                self._wbem_server = self._pywbem_server.wbem_server
            return self._pywbem_server.wbem_server
        else:
            raise click.ClickException(
                'No server specified for a command that requires a WBEM '
                'server. To specify a server, use the "--server", '
                '"--mock-server", or "--name" general options, or set the '
                'corresponding environment variables, or in interactive mode '
                'use "connection select"')

    @property
    def pywbem_server(self):
        """
        :class:`~pywbem.WBEMServer` WBEMServer instance to be used for
        requests.
        """
        return self._pywbem_server

    @property
    def spinner_enabled(self):
        """
        :class:`py:bool`: Indicates and controls whether the spinner is enabled.

        If the spinner is enabled, subcommands will display a spinning wheel
        while waiting for completion.

        This attribute can be modified.

        The initial state of the spinner is enabled, but it can be disabled by
        setting the {0} environment variable to 'false', '0', or the empty
        value.
        """.format(self.spinner_envvar)

        # Deferred initialization
        if self._spinner_enabled is None:
            value = os.environ.get(self.spinner_envvar, None)
            if value is None:
                # Default if not set
                self._spinner_enabled = True
            elif value == '0' or value == '' or value.lower() == 'false':
                self._spinner_enabled = False
            else:
                self._spinner_enabled = True

        return self._spinner_enabled

    @spinner_enabled.setter
    def spinner_enabled(self, enabled):
        """Setter method; for a description see the getter method."""
        # pylint: disable=attribute-defined-outside-init
        self._spinner_enabled = enabled

    def spinner_start(self):
        """
        Start the spinner, if the spinner is enabled.
        """
        if self.spinner_enabled:
            self._spinner_obj.start()

    def spinner_stop(self):
        """
        Stop the spinner, if the spinner is enabled.
        """
        if self.spinner_enabled:
            self._spinner_obj.stop()

    @property
    def verbose(self):
        """
        :bool:` '~click.verbose.
        """
        return self._verbose

    def set_connection(self, connection):
        """ Set the connection parameter as the current connection object and
            establish the new connection
        """
        self._pywbem_server = connection

    def execute_cmd(self, cmd):
        """
        Call the cmd executor defined by cmd with the spinner. If the
        WBEM server object has not been created it is created so that this wbem
        server can be used for interactive commands.

        This method is called by every command execution to setup and
        execute the command. Thus, each command has the line similar to:

        context.execute_cmd(lambda: cmd_instance_query(context, query, options))
        """
        # Verbose and env var PYWBEMCLI_DIAGNOSTICS displays the click
        # interpetation of the command and parameters from click context
        # This is a diagnostic for developer internal use so the
        # env variable is not documented.
        if self.verbose and os.getenv('PYWBEMCLI_DIAGNOSTICS'):
            ctx = click.get_current_context()
            click.echo('COMMAND="{}", params={!r}'.format(ctx.info_name,
                                                          ctx.params))
            display_click_context_parents(display_attrs=True)

        if not self.pdb:
            self.spinner_start()
        if self.pdb:
            import pdb  # pylint: disable=import-outside-toplevel
            pdb.set_trace()
        try:
            cmd()
        finally:
            if not self.pdb:
                self.spinner_stop()

            # Issue statistics if required. Note that we use _conn in order
            # not to create the connection if not created.
            if self.timestats and self._conn:
                click.echo(self.format_statistics(self.conn.statistics))

    def format_statistics(self, statistics):
        """
        table formatted output of statistics
        """

        def format_float3(avg, min_, max_):
            """Display float statistics with 3 places"""
            if avg == min_ == max_:
                return '{0:.3f}'.format(avg)
            return '{0:.3f}/{1:.3f}/{2:.3f}'.format(avg, min_, max_)

        def format_float0(avg, min_, max_):
            """Display float statistics with 0 places"""
            if avg == min_ == max_:
                return '{0:.0f}'.format(avg)
            return '{0:.0f}/{1:.0f}/{2:.0f}'.format(avg, min_, max_)

        snapshot = sorted(statistics.snapshot(),
                          key=lambda item: item[1].avg_time,
                          reverse=True)

        # Test to see if any server time is non-zero
        include_svr = False
        for name, stats in snapshot:  # pylint: disable=unused-variable
            # TODO: clean up pywbem stats so this is in pywbem, not here
            if stats._server_time_stored:  # pylint: disable=protected-access
                include_svr = True

        # build list of column names
        hdr = ['Count', 'Exc', 'Time']
        if include_svr:
            hdr.append('SvrTime')
        hdr.extend(['ReqLen', 'ReplyLen', 'Operation'])

        # build table rows from snapshot of OperationStatistics
        rows = []
        for name, stats in snapshot:  # pylint: disable=unused-variable
            time = format_float3(stats.avg_time,
                                 stats.min_time,
                                 stats.max_time)
            req_len = format_float0(stats.avg_request_len,
                                    stats.min_request_len,
                                    stats.max_request_len)
            reply_len = format_float0(stats.avg_reply_len,
                                      stats.min_reply_len,
                                      stats.max_reply_len)
            if include_svr:
                svrtime = format_float3(stats.avg_server_time,
                                        stats.min_server_time,
                                        stats.max_server_time)
                row = [stats.count, stats.exception_count, time, svrtime,
                       req_len, reply_len, stats.name]
            else:
                row = [stats.count, stats.exception_count, time,
                       req_len, reply_len, stats.name]
            rows.append(row)

        # only add table description if verbose on.
        if self.verbose:
            title = 'Statistics: Time in sec. Time, ReqLen, etc are single ' \
                    'value if average/min/max are the same'
        else:
            title = None

        click.echo(format_table(rows, hdr, title=title))

    def connect_wbem_server(self):
        """
        If the wbem server has not been connected yet, connect it. The
        wbemserver object is saved both in the context and as a GLOBAL
        This GLOBAL is used to determine that the server has been connected.
        TODO: FUTURE Rather than a global, we should be the wbemserver into the
        parent probably.

        Both the WBEMConnection and WBEMServer objects are established
        at this point since the cost of doing the WBEMServer object is low.
        The existence of the WBEMServer object in the GLOBALS is the flag that
        this server connection has been established already.

        TODO: FUTURE - We should probably move this logic to the first place a
        connection is actually established so that if we have not-password
        commands in the environment they do not force the password request.
        Logical would be the first time the conn or wbem_server property is
        used in pywbem_server.

        """
        # TODO: FUTURE investigate putting this into parent context

        # If no server defined, do not try to connect. This allows
        # commands like help, connection new, select to execute without
        # a target server defined.
        if self._pywbem_server and self._pywbem_server.wbem_server is None:
            # get the password if it is required.  This may involve a
            # prompt.
            self._pywbem_server.get_password(self)
            self._pywbem_server.create_connection(
                log=self.log,
                use_pull=self.use_pull,
                timestats=self.timestats,
                verbose=self.verbose)

    @staticmethod
    def update_root_click_context(ctx_obj):
        """
        Method to update the click root context with a new context
        object.  This makes the values in this context universal for all
        future commands within an interactive session.
        This is a static method and gets the current click context and
        root context from click itself.
        """
        ctx = click.get_current_context()
        root = ctx.find_root()
        root.obj = ctx_obj


#
#   Debug tools to help understandingthe click context.  These are only
#   to help analyzing the click_context (ctx) and primarily in the
#   interactive mode whene several levels of click context can exist
#
def display_click_context(ctx, msg=None, display_attrs=True):
    """
    Debug function displays attributes of click context
    This is a diagnostic for developer use
    """

    attrs = vars(ctx)
    if not msg:
        msg = "CLICK_CONTEXT"
    if not display_attrs:
        click.echo(ctx.obj)
    else:
        click.echo('{0} {1}, attrs: {2}'.format(
            msg, ctx,
            '\n    '.join('{0}: {1}'.format(i, v) for (i, v) in attrs.items())))


def display_click_context_parents(display_attrs=False):
    """
    Display the current click context and its all of its parents.
    This is a diagnostic for developer use
    """
    level = 0
    ctx = click.get_current_context()
    disp_ctx = ctx.parent
    while disp_ctx is not None:
        display_click_context(disp_ctx, msg='level {}'.format(level),
                              display_attrs=display_attrs)
        level += -1
        disp_ctx = disp_ctx.parent



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pywbemcli is a command line WBEM client that uses pywbem as its communication
interface with WBEM Servers. It is written in pure Python and supports Python 2
and Python 3.
"""

from __future__ import absolute_import, print_function

from ._cmd_class import *           # noqa: F403,F401
from ._cmd_instance import *        # noqa: F403,F401
from ._cmd_qualifier import *       # noqa: F403,F401
from ._cmd_server import *          # noqa: F403,F401
from ._cmd_connection import *      # noqa: F403,F401
from ._common import *              # noqa: F403,F401
from ._pywbem_server import *       # noqa: F403,F401
from ._context_obj import *         # noqa: F403,F401
from ._connection_repository import *   # noqa: F403,F401
from .pywbemcli import *            # noqa: F403,F401
from .config import *               # noqa: F403,F401
from ._pywbemcli_operations import *  # noqa: F403,F401
from ._click_extensions import *    # noqa: F403,F401
from ._association_shrub import *   # noqa: F403,F401
from ._utils import *               # noqa: F403,F401
from ._cimvalueformatter import *   # noqa: F403,F401


from .._version import __version__  # noqa: F401



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click Command definition for the instance command group which includes
cmds for get, enumerate, list of instance.

NOTE: Commands are ordered in help display by their order in this file.
"""

from __future__ import absolute_import, print_function

import click

from pywbem import CIMInstanceName, CIMClassName, Error, CIMError, \
    CIM_ERR_NOT_FOUND

from .pywbemcli import cli
from ._common import display_cim_objects, \
    pick_instance, resolve_propertylist, create_ciminstance, \
    filter_namelist, format_table, verify_operation, \
    process_invokemethod, raise_pywbem_error_exception, \
    parse_kv_pair, warning_msg, validate_output_format, \
    CMD_OPTS_TXT, GENERAL_OPTS_TXT, SUBCMD_HELP_TXT

from ._common_options import add_options, propertylist_option, \
    names_only_option, include_classorigin_instance_option, namespace_option, \
    summary_option, verify_option, multiple_namespaces_option, \
    association_filter_option, indication_filter_option, \
    experimental_filter_option, help_option

from ._cimvalueformatter import mof_escaped

from ._association_shrub import AssociationShrub

from .config import DEFAULT_QUERY_LANGUAGE
from ._click_extensions import PywbemcliGroup, PywbemcliCommand
from ._cmd_class import get_namespaces, enumerate_classes_filtered

#
#   Common option definitions for instance group
#

# NOTE: A number of the options use double-dash as the short form.  In those
# cases, a third definition of the options without the double-dash defines
# the corresponding option name, ex. 'include_qualifiers'. It should be
# defined with underscore and not dash

# Issue 224 - Exception in prompt-toolkit with python 2.7. Caused because
# with prompt-toolkit 2 + the completer requires unicode and click_repl not
# passing help as unicode in options as unicode
# NOTE: Insure that all option help attributes are unicode to get around this
#       issue


# This is instance-only because the default is False for include-qualifiers
# on instances but True on classes
include_qualifiers_get_option = [              # pylint: disable=invalid-name
    click.option('--iq', '--include-qualifiers', 'include_qualifiers',
                 is_flag=True, required=False,
                 help=u'Include qualifiers in the returned instance. '
                      u'Not all servers return qualifiers on instances. '
                      u'Default: Do not include qualifiers.')]

include_qualifiers_list_option = [              # pylint: disable=invalid-name
    click.option('--iq', '--include-qualifiers', 'include_qualifiers',
                 is_flag=True, required=False,
                 help=u'When traditional operations are used, include '
                      u'qualifiers in the returned instances. '
                      u'Some servers may ignore this option. '
                      u'By default, and when pull operations are used, '
                      u'qualifiers will never be included.')]

# specific to instance because DeepInheritance differs between class and
# instance operations.
deep_inheritance_enum_option = [              # pylint: disable=invalid-name
    click.option('--di', '--deep-inheritance', 'deep_inheritance',
                 is_flag=True, required=False,
                 help=u'Include subclass properties in the returned '
                      u'instances. '
                      u'Default: Do not include subclass properties.')]

local_only_get_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 required=False,
                 help=u'Do not include superclass properties in the returned '
                      u'instance. '
                      u'Some servers may ignore this option. '
                      u'Default: Include superclass properties.')]

local_only_list_option = [              # pylint: disable=invalid-name
    click.option('--lo', '--local-only', 'local_only', is_flag=True,
                 required=False,
                 help=u'When traditional operations are used, do not include '
                      u'superclass properties in the returned instances. '
                      u'Some servers may ignore this option. '
                      u'By default, and when pull operations are used, '
                      u'superclass properties will always be included.')]

property_create_option = [              # pylint: disable=invalid-name
    click.option('-p', '--property', type=str, metavar='PROPERTYNAME=VALUE',
                 required=False, multiple=True,
                 help=u'Initial property value for the new instance. '
                      u'May be specified multiple times. '
                      u'Array property values are specified as a '
                      u'comma-separated list; embedded instances are not '
                      u'supported. '
                      u'Default: No initial properties provided.')]

property_modify_option = [              # pylint: disable=invalid-name
    click.option('-p', '--property', type=str, metavar='PROPERTYNAME=VALUE',
                 required=False, multiple=True,
                 help=u'Property to be modified, with its new value. '
                      u'May be specified once for each property to be '
                      u'modified. '
                      u'Array property values are specified as a '
                      u'comma-separated list; embedded instances are not '
                      u'supported. '
                      u'Default: No properties modified.')]

keybinding_key_option = [              # pylint: disable=invalid-name
    click.option('-k', '--key', type=str, metavar='KEYNAME=VALUE',
                 required=False, multiple=True,
                 help=u'Value for a key in keybinding of CIM instance name. '
                      u'May be specified multiple times. '
                      u'Allows defining keys without the issues of quotes. '
                      u'Default: No keybindings provided.')]

filter_query_language_option = [              # pylint: disable=invalid-name
    click.option('--fql', '--filter-query-language', 'filter_query_language',
                 type=str, metavar='QUERY-LANGUAGE', default=None,
                 help=u'The filter query language to be used with '
                      u'--filter-query. '
                      u'Default: DMTF:FQL.')]

filter_query_option = [              # pylint: disable=invalid-name
    click.option('--fq', '--filter-query', 'filter_query', type=str,
                 metavar='QUERY-STRING', default=None,
                 help=u'When pull operations are used, filter the instances in '
                      u'the result via a filter query. '
                      u'By default, and when traditional operations are used, '
                      u'no such filtering takes place.')]

help_instancename_option = [              # pylint: disable=invalid-name
    click.option('--hi', '--help-instancename', 'help_instancename',
                 is_flag=True, required=False, is_eager=True,
                 help=u'Show help message for specifying INSTANCENAME '
                      u'including use of the --key and --namespace options.')]


##########################################################################
#
#   Click command group and command definitions
#   These decorated functions implement the commands, arguments, and
#
###########################################################################


@cli.group('instance', cls=PywbemcliGroup, options_metavar=GENERAL_OPTS_TXT,
           subcommand_metavar=SUBCMD_HELP_TXT)
@add_options(help_option)
def instance_group():
    """
    Command group for CIM instances.

    This command group defines commands to inspect instances, to invoke
    methods on instances, and to create and delete instances.

    Modification of instances is not currently supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'pywbemcli --help') can also be specified before the
    'instance' keyword.
    """
    pass  # pylint: disable=unnecessary-pass


@instance_group.command('enumerate', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@add_options(local_only_list_option)
@add_options(deep_inheritance_enum_option)
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(namespace_option)
@add_options(names_only_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@add_options(help_option)
@click.pass_obj
def instance_enumerate(context, classname, **options):
    """
    List the instances of a class.

    Enumerate the CIM instances of the specified class (CLASSNAME argument),
    including instances of subclasses in the specified CIM namespace
    (--namespace option), and display the returned instances, or instance paths
    if --names-only was specified. If no namespace was specified, the default
    namespace of the connection is used.

    The instances to be retrieved can be filtered by the --filter-query option.

    The --local-only, --deep-inheritance, --include-qualifiers,
    --include-classorigin, and --propertylist options determine which parts
    are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    context.execute_cmd(lambda: cmd_instance_enumerate(context, classname,
                                                       options))


@instance_group.command('get', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@add_options(local_only_get_option)
@add_options(include_qualifiers_get_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_get(context, instancename, **options):
    """
    Get an instance of a class.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    The --local-only, --include-qualifiers, --include-classorigin, and
    --propertylist options determine which parts are included in the retrieved
    instance.

    In the output, the instance will formatted as defined by the
    --output-format general option.
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_get(context, instancename,
                                                 options))


@instance_group.command('delete', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_delete(context, instancename, **options):
    """
    Delete an instance of a class.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_delete(context, instancename,
                                                    options))


@instance_group.command('create', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME', required=True)
@add_options(property_create_option)
@add_options(verify_option)
@add_options(namespace_option)
@add_options(help_option)
@click.pass_obj
def instance_create(context, classname, **options):
    """
    Create an instance of a class in a namespace.

    Create a CIM instance of the specified creation class (CLASSNAME
    argument) in the specified CIM namespace (--namespace option), with
    the specified properties (--property options) and display the CIM instance
    path of the created instance. If no namespace was specified, the default
    namespace of the connection is used.

    The properties to be initialized and their new values are specified using
    the --property option, which may be specified multiple times.

    Pywbemcli retrieves the class definition from the server in order to
    verify that the specified properties are consistent with the property
    characteristics in the class definition.

    Example:

      pywbemcli instance create CIM_blah -P id=3 -P arr="bla bla",foo
    """
    context.execute_cmd(lambda: cmd_instance_create(context, classname,
                                                    options))


@instance_group.command('modify', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@add_options(property_modify_option)
@click.option('--pl', '--propertylist', 'propertylist', multiple=True, type=str,
              default=None, required=False, metavar='PROPERTYLIST',
              help=u'Reduce the properties to be modified (as per '
              u'--property) to a specific property list. '
              u'Multiple properties may be specified with either a '
              u'comma-separated list or by using the option multiple '
              u'times. The empty string will cause no properties to '
              u'be modified. '
              u'Default: Do not reduce the properties to be modified.')
@add_options(verify_option)
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_modify(context, instancename, **options):
    """
    Modify properties of an instance.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    The properties to be modified and their new values are specified using the
    --property option, which may be specified multiple times.

    The --propertylist option allows restricting the set of properties to be
    modified. Given that the set of properties to be modified is already
    determined by the specified --property options, it does not need to be
    specified.

    Example:

      pywbemcli instance modify CIM_blah.fred=3 -P id=3 -P arr="bla bla",foo
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_modify(context, instancename,
                                                    options))


@instance_group.command('associators', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help=u'Filter the result set by association class name. '
                   u'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   u'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by far end role name.')
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_associators(context, instancename, **options):
    """
    List the instances associated with an instance.

    List the CIM instances that are associated with the specified CIM instance,
    and display the returned instances, or instance paths if --names-only was
    specified.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    The instances to be retrieved can be filtered by the --filter-query,
    --role, --result-role, --assoc-class, and --result-class options.

    The --include-qualifiers, --include-classorigin, and --propertylist options
    determine which parts are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_associators(context, instancename,
                                                         options))


@instance_group.command('references', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   u'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False, metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@add_options(include_qualifiers_list_option)
@add_options(include_classorigin_instance_option)
@add_options(propertylist_option)
@add_options(names_only_option)
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(summary_option)
@add_options(filter_query_option)
@add_options(filter_query_language_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_references(context, instancename, **options):
    """
    List the instances referencing an instance.

    List the CIM (association) instances that reference the specified CIM
    instance, and display the returned instances, or instance paths if
    --names-only was specified.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    The instances to be retrieved can be filtered by the --filter-query, --role
    and --result-class options.

    The --include-qualifiers, --include-classorigin, and --propertylist options
    determine which parts are included in each retrieved instance.

    The --names-only option can be used to show only the instance paths.

    In the output, the instances and instance paths will be formatted as
    defined by the --output-format general option. Table formats on instances
    will be replaced with MOF format.
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_references(context, instancename,
                                                        options))


@instance_group.command('invokemethod', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@click.argument('methodname', type=str, metavar='METHODNAME', required=False)
@click.option('-p', '--parameter', type=str, metavar='PARAMETERNAME=VALUE',
              required=False, multiple=True,
              help=u'Specify a method input parameter with its value. '
                   u'May be specified multiple times. '
                   u'Array property values are specified as a comma-separated '
                   u'list; embedded instances are not supported. '
                   u'Default: No input parameters.')
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_invokemethod(context, instancename, methodname, **options):
    """
    Invoke a method on an instance.

    Invoke a CIM method (METHODNAME argument) on a CIM instance with the
    specified input parameters (--parameter options), and display the method
    return value and output parameters.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    The method input parameters are specified using the --parameter option,
    which may be specified multiple times.

    Pywbemcli retrieves the class definition of the creation class of the
    instance from the server in order to verify that the specified input
    parameters are consistent with the parameter characteristics in the method
    definition.

    Use the 'class invokemethod' command to invoke CIM methods on CIM classes.

    Example:

      pywbemcli -n myconn instance invokemethod CIM_x.id='hi" methodx -p id=3
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    validate_required_arg(methodname, 'METHODNAME')
    context.execute_cmd(lambda: cmd_instance_invokemethod(context,
                                                          instancename,
                                                          methodname,
                                                          options))


@instance_group.command('query', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('query', type=str, required=True, metavar='QUERY-STRING')
@click.option('--ql', '--query-language', 'query_language', type=str,
              metavar='QUERY-LANGUAGE', default=DEFAULT_QUERY_LANGUAGE,
              help=u'The query language to be used with --query. '
              u'Default: {default}.'.
              format(default=DEFAULT_QUERY_LANGUAGE))
@add_options(namespace_option)
@add_options(summary_option)
@add_options(help_option)
@click.pass_obj
def instance_query(context, query, **options):
    """
    Execute a query on instances in a namespace.

    Execute the specified query (QUERY_STRING argument) in the specified CIM
    namespace (--namespace option), and display the returned instances. If no
    namespace was specified, the default namespace of the connection is used.

    In the output, the instances will formatted as defined by the
    --output-format general option.
    """
    context.execute_cmd(lambda: cmd_instance_query(context, query, options))


@instance_group.command('count', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('classname', type=str, metavar='CLASSNAME-GLOB',
                required=False)
@add_options(multiple_namespaces_option)
@add_options(association_filter_option)
@add_options(indication_filter_option)
@add_options(experimental_filter_option)
@click.option('-s', '--sort', is_flag=True, required=False,
              help=u'Sort by instance count. Otherwise sorted by class name.')
@add_options(help_option)
@click.pass_obj
def instance_count(context, classname, **options):
    """
    Count the instances of each class with matching class name.

    Display the count of instances of each CIM class whose class name
    matches the specified wildcard expression (CLASSNAME-GLOB) in all CIM
    namespaces of the WBEM server, or in the specified namespaces
    (--namespace option).  This differs from instance enumerate, etc. in that
    it counts the instances specifically for the classname of each instance
    returned, not including subclasses.

    The CLASSNAME-GLOB argument is a wildcard expression that is matched on
    class names case insensitively.
    The special characters from Unix file name wildcarding are supported
    ('*' to match zero or more characters, '?' to match a single character,
    and '[]' to match character ranges). To avoid shell expansion of wildcard
    characters, the CLASSNAME-GLOB argument should be put in quotes.

    If CLASSNAME-GLOB is not specified, then all classes in the specified
    namespaces are counted (same as when specifying CLASSNAME-GLOB as "*").

    For example, "pywbem_*" returns classes whose name begins with "PyWBEM_",
    "pywbem_", etc. "*system*" returns classes whose names include the case
    insensitive string "system".

    This command can take a long time to execute since it potentially
    enumerates all instance names for all classes in all namespaces.
    """
    context.execute_cmd(lambda: cmd_instance_count(context, classname, options))


@instance_group.command('shrub', cls=PywbemcliCommand,
                        options_metavar=CMD_OPTS_TXT)
@click.argument('instancename', type=str, metavar='INSTANCENAME',
                required=False)
@click.option('--ac', '--assoc-class', 'assoc_class', type=str, required=False,
              metavar='CLASSNAME',
              help=u'Filter the result set by association class name. '
                   'Subclasses of the specified class also match.')
@click.option('--rc', '--result-class', 'result_class', type=str,
              required=False, metavar='CLASSNAME',
              help=u'Filter the result set by result class name. '
                   u'Subclasses of the specified class also match.')
@click.option('-r', '--role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by source end role name.')
@click.option('--rr', '--result-role', 'result_role', type=str, required=False,
              metavar='PROPERTYNAME',
              help=u'Filter the result set by far end role name.')
@add_options(keybinding_key_option)
@add_options(namespace_option)
@add_options(summary_option)
@click.option('-f', '--fullpath', default=False, is_flag=True,
              help=u'Normally the instance paths in the tree views are '
                   u'by hiding some keys with ~ to make the tree simpler '
                   u'to read. This includes keys that have the same value '
                   u'for all instances and the "CreationClassName" key.  When'
                   'this option is used the full instance paths are displayed.')
@add_options(help_instancename_option)
@add_options(help_option)
@click.pass_obj
def instance_shrub(context, instancename, **options):
    """
    Show the association shrub for INSTANCENAME.

    The shrub is a view of all of the instance association relationships for
    a defined INSTANCENAME showing the various components that are part of
    the association including Role, AssocClasse,ResultRole, And ResultClas

    The default view is a tree view from the INSTANCENAME to associated
    instances.

    Displays the shrub of association components for the association source
    instance defined by INSTANCENAME.

    For information on how to specify the instance using INSTANCENAME and the
    --key and --namespace options, invoke with --help-instancename.

    Normally the association information is displayed as a tree but it
    may also be displayed as a table or as one of the object formats (ex. MOF)
    of all instances that are part of the shrub if one of the cim object
    formats is selected with the global output_format parameter.

    Results are formatted as defined by the output format global option.
    """
    if options['help_instancename']:
        show_help_instancename()
        return
    validate_required_arg(instancename, 'INSTANCENAME')
    context.execute_cmd(lambda: cmd_instance_shrub(context, instancename,
                                                   options))


####################################################################
#
#  Common functions for cmd_instance processing
#
####################################################################


def validate_required_arg(value, metavar):
    """
    Validate that a required CLI argument is present, and raise a usage error
    otherwise.
    """
    if not value:
        raise click.UsageError(
            "Missing argument '{}'.".format(metavar),
            click.get_current_context())


def show_help_instancename():
    """
    Show the help message on how to specify an instance using INSTANCENAME
    and the --key and --namespace options.
    """
    # TODO: Reflow according to target width
    click.echo("""
An instance path is specified using the INSTANCENAME argument and optionally the
--key and --namespace options. It can be specified in three ways:

1. By specifying the instance path as an untyped WBEM URI in the INSTANCENAME
   argument. In this case, the keybindings of the instance are specified in the
   WBEM URI, and the --key option must not be used.

   String typed keybindings in WBEM URIs are using double quotes, so the
   processing of special characters by the shell (such as backslashes and
   quotes) needs to be considered.

   The CIM namespace of the instance can be specified in the WBEM URI, or using
   the --namespace option, or otherwise the default namespace of the connection
   will be used.

   If present, the scheme and host components in the WBEM URI will be ignored.

   Examples (for a shell such as 'bash'):

     cimv2/test:TST_Person.FirstName=\\"Albert\\",LastName=\\"Einstein\\"
     TST_Foo.ID=42 --namespace cimv2/test
     TST_Foo.ID=42                     # default namespace of connection is used
     TST_Foo.IntegerKey=42
     TST_Foo.BooleanKey=true
     TST_Foo.StringKey=\\"text\\"        # shell processes escapes
     'TST_Foo.StringKey="text"'        # shell processes outer squotes
     'TST_Foo.StringKey="42"'          # shell processes outer squotes
     'TST_Foo.StringKey="true"'        # shell processes outer squotes
     "TST_Foo.StringKey=\\"it's\\""      # shell proc. outer dquotes and escapes
     'TST_Foo.StringKey="1.75\\""'      # shell processes outer squotes
     'TST_Foo.StringKey="a b"'         # shell processes outer squotes
     'CIM_SubProfile.Main="CIM_RegisteredProfile.InstanceID=\\"acme.1\\"",
       Sub="CIM_RegisteredProfile.InstanceID=\\"acme.2\\""'

2. By specifying the class path of the creation class of the instance as an
   untyped WBEM URI in the INSTANCENAME argument and by using the --key option
   to specify the keybindings of the instance.

   This approach reduces the use of double quotes compared to the first
   approach and eliminates it for the most common cases, but the processing of
   special characters by the shell still needs to be considered.

   The --key option can be specified multiple times, once for each key of the
   instance name. The argument of the --key option has a NAME=VALUE format,
   where NAME is the name of the key property and VALUE is its value. The
   string/numeric/boolean type needed for creating a keybinding is determined
   automatically from VALUE. Valid integer numbers are interpreted as a numeric
   type, the strings "true" and "false" in any lexical case are interpreted as
   a boolean type, and anything else is interpreted as a string type. Starting
   and ending VALUE with double quotes forces interpretation as a string type;
   in that case double quotes and backslashes inside of the double quotes need
   to be backslash-escaped. This is useful for strings that have a numeric
   value, or the string values "true" or "false".

   The CIM namespace of the instance can be specified in the WBEM URI, or using
   the --namespace option, or otherwise the default namespace of the connection
   will be used.

   If present, the scheme and host components in the WBEM URI will be ignored.

   Examples (for a shell such as 'bash'):

     cimv2/test:TST_Person --key FirstName=Albert --key LastName=Einstein
     TST_Foo --namespace cimv2/test --key ID=42
     TST_Foo --key ID=42               # default namespace of connection is used
     TST_Foo --key IntegerKey=42
     TST_Foo --key BooleanKey=true
     TST_Foo --key StringKey=text
     TST_Foo --key StringKey=\\"42\\"    # shell processes escapes
     TST_Foo --key StringKey=\\"true\\"  # shell processes escapes
     TST_Foo --key "StringKey=it's"    # shell processes outer dquotes
     TST_Foo --key 'StringKey=1.75"'   # shell processes outer squotes
     TST_Foo --key "StringKey=a b"     # shell processes outer dquotes
     TST_Foo --key StringKey="a b"     # shell processes outer dquotes
     TST_Foo --key StringKey=a\\ b      # shell processes escapes
     CIM_SubProfile --key Main='CIM_RegisteredProfile.InstanceID="acme.1"'
       --key Sub='CIM_RegisteredProfile.InstanceID="acme.2"'

3. By specifying the class path of the creation class of the instance as an
   untyped WBEM URI in the INSTANCENAME argument, followed by a wildcard
   indicator ".?". In this case, the --key option must not be used. The
   instances of the specified class are displayed and the user is prompted for
   an index number to select an instance.

   The CIM namespace of the instance can be specified in the WBEM URI, or using
   the --namespace option, or otherwise the default namespace of the connection
   will be used.

   If present, the scheme and host components in the WBEM URI will be ignored.

   Example command:

     $ pywbemcli instance get cimv2/test:TST_Person.?
     Pick Instance name to process
     0: cimv2/test:TST_Person.FirstName="Albert",LastName="Einstein"
     1: cimv2/test:TST_Person.FirstName="Marie",LastName="Curie"
     Input integer between 0 and 1 or Ctrl-C to exit selection: _
""")


def get_instancename(context, instancename, options):
    """
    Common function to construct a CIMInstanceName object from the
    INSTANCENAME argument and the --key and --namespace options specified
    in the command line.

    The keybindings of the returned CIM instance path must be specified in
    exactly one of these ways:

    * If the keybindings component in the instance name string is "?"
      (e.g. "CIM_Foo.?"), the instances of that class are listed and the user is
      prompted to pick one.

    * If the "key" option is non-empty, the so specified key values are used as
      keybindings for the returned instance path, and the instance name string
      is interpreted as a class WBEM URI.

    * If the instance name string specifies keybindings, they are used.

    The namespace of the returned CIM instance path must be specified in
    exactly one of these ways:

    * If the instance name string specifies a namespace, it is used.

    * If the "namespace" option is non-empty, it is used.

    * Otherwise, the default namespace of the connection in the context is used.

    Parameters:

      context (:class:`'~pywbemtools._context_obj.ContextObj`):
        The Click context object.

      instancename (:term:`string`):
        The INSTANCENAME argument from the command line.

      options (dict):
        Command-specific options from the command line (including --key and
        --namespace if specified).

    Returns:

      :class:`~pywbem.CIMInstanceName`: CIM instance path. It is never None.

    Raises:

      ClickException: Various reasons.
    """

    if instancename.endswith(".?"):

        if options['key']:
            raise click.ClickException(
                "Using the --key option conflicts with specifying a "
                "wildcard keybinding in INSTANCENAME: {}".
                format(instancename))

        class_uri = instancename[:-2]
        try:
            class_path = CIMClassName.from_wbem_uri(class_uri)
        except ValueError as exc:
            raise click.ClickException(str(exc))

        if class_path.namespace:
            if options.get('namespace'):
                raise click.ClickException(
                    "Using the --namespace option conflicts with specifying a "
                    "namespace in INSTANCENAME: {}".format(instancename))
        else:
            class_path.namespace = options.get('namespace') or \
                context.conn.default_namespace

        try:
            instance_path = pick_instance(
                context, class_path.classname, class_path.namespace)
        except ValueError as exc:
            raise click.ClickException(str(exc))

    elif options['key']:

        # Transform the --key option values into WBEM URI keybinding strings
        kb_strs = []
        for kv in options['key']:
            key, value = parse_kv_pair(kv)
            if value is None:
                raise click.ClickException(
                    "VALUE in --key option argument is missing: {}".format(kv))
            try:
                int(value)
                is_int = True
            except (ValueError, TypeError):
                is_int = False
            if value is None:
                # There is no official NULL representation in WBEM URI
                kb_value = ''
            elif is_int:
                # integer - use without quotes
                kb_value = value
            elif value.upper() in ('TRUE', 'FALSE'):
                # boolean - use the upper cased string without quotes
                kb_value = value.upper()
            elif value.startswith('"') and value.endswith('"'):
                # string - the value between the double quotes is assumed to
                # already be backslash-escaped, at the minimum for double
                # quotes and backslashes.
                kb_value = value
            else:
                # string - double quotes and escaping is added
                # Note that a keybinding value in a WBEM URI only requires
                # the minimal escaping for MOF string literals, i.e. double
                # quotes and backslashes. However, we escape all control chars,
                # double quotes, and single quotes, to ensure that there are no
                # unprintable characters, and no quotes that might interfere
                # with the shell.
                kb_value = '"{}"'.format(mof_escaped(value))
            kb_strs.append("{}={}".format(key, kb_value))

        # We perform an extra verification that instancename was a class path,
        # in order to get a more understandable error message if it is not,
        # compared to leaving that to CIMInstanceName.from_wbem_uri().
        try:
            CIMClassName.from_wbem_uri(instancename)
        except ValueError as exc:
            raise click.ClickException(str(exc))

        instancename_kb = "{}.{}".format(instancename, ','.join(kb_strs))

        try:
            instance_path = CIMInstanceName.from_wbem_uri(instancename_kb)
        except ValueError as exc:
            raise click.ClickException(str(exc))

        if instance_path.namespace:
            if options.get('namespace'):
                raise click.ClickException(
                    "Using the --namespace option conflicts with specifying a "
                    "namespace in INSTANCENAME: {}".format(instancename))
        else:
            instance_path.namespace = options.get('namespace') or \
                context.conn.default_namespace

    else:

        assert not options['key']  # There cannot be a conflict anymore

        try:
            instance_path = CIMInstanceName.from_wbem_uri(instancename)
        except ValueError as exc:
            raise click.ClickException(str(exc))

        if instance_path.namespace:
            if options.get('namespace'):
                raise click.ClickException(
                    "Using the --namespace option conflicts with specifying a "
                    "namespace in INSTANCENAME: {}".format(instancename))
        else:
            instance_path.namespace = options.get('namespace') or \
                context.conn.default_namespace

    return instance_path


####################################################################
#
#  cmd_instance_<action> processors
#
####################################################################


def cmd_instance_get(context, instancename, options):
    """
    Get and display an instance of a CIM Class.

    Gets the instance defined by instancename argument and displays in output
    format defined.

    If the wildcard key is used (CLASSNAME.?), pywbemcli presents a list of
    instances to the console from which one can be picked to get from the
    server and display.
    """
    output_fmt = validate_output_format(context.output_format, ['CIM', 'TABLE'])

    instancepath = get_instancename(context, instancename, options)

    try:
        instance = context.conn.GetInstance(
            instancepath,
            LocalOnly=options['local_only'],
            IncludeQualifiers=options['include_qualifiers'],
            IncludeClassOrigin=options['include_classorigin'],
            PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, instance, output_fmt)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_instance_delete(context, instancename, options):
    """
        If option interactive is set, get instances of the class defined
        by instance name and allow the user to select the instance to
        delete.
        Otherwise attempt to delete the instance defined by instancename
    """
    instancepath = get_instancename(context, instancename, options)

    try:
        context.conn.DeleteInstance(instancepath)

        if context.verbose:
            context.spinner_stop()
            click.echo('Deleted instance {}'.format(instancepath))

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_instance_create(context, classname, options):
    """
       Create an instance and submit to wbemserver.
       If successful, this operation returns the new instance name. Otherwise
       it raises an exception
    """
    ns = options['namespace'] or context.conn.default_namespace
    try:
        class_ = context.conn.GetClass(
            classname, namespace=ns, LocalOnly=False)
    except CIMError as ce:
        if ce.status_code == CIM_ERR_NOT_FOUND:
            raise click.ClickException('CIMClass: "{}" does not exist in '
                                       'namespace "{}" in WEB '
                                       'server: {}.'.format(classname, ns,
                                                            context.conn))
        raise_pywbem_error_exception(ce)

    except Error as er:
        raise_pywbem_error_exception(er)

    properties = options['property']

    # properties is a tuple of name,value pairs
    new_inst = create_ciminstance(class_, properties)

    if options['verify']:
        context.spinner_stop()
        click.echo(new_inst.tomof())
        if not verify_operation("Execute CreateInstance", msg=True):
            return
    try:
        name = context.conn.CreateInstance(new_inst,
                                           namespace=ns)

        context.spinner_stop()
        click.echo('{}'.format(name))
    except Error as er:
        raise click.ClickException('Server Error creating instance in '
                                   'namespace {}. Exception: '
                                   '{}: {}'.format(ns, er.__class__.__name__,
                                                   er))


def cmd_instance_modify(context, instancename, options):
    """
    Build an instance defined by the options and submit to wbemserver
    as a ModifyInstance method.

    In order to make a correct instance, this method first gets the
    corresponding class and uses that as the template for creating the intance.
    The class provides property names, types, array flags, etc. to assure
    that the instance is correctly built.

    If successful, this operation returns nothing.
    """
    instancepath = get_instancename(context, instancename, options)

    ns = options['namespace'] or context.conn.default_namespace

    try:
        class_ = context.conn.GetClass(
            instancepath.classname, namespace=ns, LocalOnly=False)
    except CIMError as ce:
        if ce.status_code == CIM_ERR_NOT_FOUND:
            raise click.ClickException(
                'CIMClass: {!r} does not exist in WEB server: {}'
                .format(instancepath.classname, context.conn.url))

        raise_pywbem_error_exception(ce)
    except Error as er:
        raise_pywbem_error_exception(er)

    property_list = resolve_propertylist(options['propertylist'])

    # properties is a tuple of name,value pairs
    modified_inst = create_ciminstance(class_, options['property'])

    modified_inst.path = instancepath

    if options['verify']:
        context.spinner_stop()
        click.echo(modified_inst.tomof())
        if not verify_operation("Execute ModifyInstance", msg=True):
            return

    try:
        context.conn.ModifyInstance(modified_inst,
                                    PropertyList=property_list)
        if context.verbose:
            context.spinner_stop()
            click.echo('Modified instance {}'.format(instancepath))
    except Error as er:
        raise click.ClickException('Server Error modifying instance {} '
                                   'in namespace {}. Exception: {}: {}'.format
                                   (instancepath, ns, er.__class__.__name__,
                                    er))


def cmd_instance_invokemethod(context, instancename, methodname,
                              options):
    """
    Create an instance and submit to wbemserver
    """
    instancepath = get_instancename(context, instancename, options)

    try:
        process_invokemethod(context, instancepath, methodname, options)
    except Error as er:
        raise_pywbem_error_exception(er)


def get_filterquerylanguage(options):
    """
    Get the filterquery language based on what is in the filterquery option
    and the filterquerylanguage options.
    If filterquery exists but filterquerylanguage does not, use DMTF as
    the filter query language.
    if filterquery does not exist but filterquerylanguage does, just return it
    """
    if options['filter_query']:
        fql = options['filter_query_language'] or 'DMTF:FQL'
    else:
        fql = options['filter_query_language']
    return fql


def cmd_instance_enumerate(context, classname, options):
    """
    Enumerate CIM instances or CIM instance names

    """
    output_fmt = validate_output_format(context.output_format, ['CIM', 'TABLE'])

    try:
        if options['names_only']:
            results = context.conn.PyWbemcliEnumerateInstancePaths(
                ClassName=classname,
                namespace=options['namespace'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliEnumerateInstances(
                ClassName=classname,
                namespace=options['namespace'],
                LocalOnly=options['local_only'],
                IncludeQualifiers=options['include_qualifiers'],
                DeepInheritance=options['deep_inheritance'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_fmt,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)
    except ValueError as ve:
        raise click.ClickException('instance enumerate failed because '
                                   'FilterQuery not allowed with traditional '
                                   'EnumerateInstance. --use-pull: '
                                   '{}. Exception: {}: {}'
                                   .format(context.use_pull,
                                           ve.__class__.__name__, ve))


def cmd_instance_references(context, instancename, options):
    """Execute the references request operation to get references for
       the classname defined. This may be either interactive or if the
       interactive option is set or use the instancename directly.

       If the interactive option is selected, the instancename MUST BE
       a classname.
    """
    output_fmt = validate_output_format(context.output_format, ['CIM', 'TABLE'])

    instancepath = get_instancename(context, instancename, options)

    try:
        if options['names_only']:
            results = context.conn.PyWbemcliReferenceInstancePaths(
                instancepath,
                ResultClass=options['result_class'],
                Role=options['role'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliReferenceInstances(
                instancepath,
                ResultClass=options['result_class'],
                Role=options['role'],
                IncludeQualifiers=options['include_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_fmt,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)
    except ValueError as ve:
        raise click.ClickException('instance references failed because '
                                   'FilterQuery not allowed with traditional '
                                   'References. --use-pull: '
                                   '{}. Exception: {}: {}'
                                   .format(context.use_pull,
                                           ve.__class__.__name__, ve))


def cmd_instance_associators(context, instancename, options):
    """
    Execute the references request operation to get references for
    the classname defined
    """
    output_fmt = validate_output_format(context.output_format, ['CIM', 'TABLE'])

    instancepath = get_instancename(context, instancename, options)

    try:
        if options['names_only']:
            results = context.conn.PyWbemcliAssociatorInstancePaths(
                instancepath,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt)
        else:
            results = context.conn.PyWbemcliAssociatorInstances(
                instancepath,
                AssocClass=options['assoc_class'],
                Role=options['role'],
                ResultClass=options['result_class'],
                ResultRole=options['result_role'],
                IncludeQualifiers=options['include_qualifiers'],
                IncludeClassOrigin=options['include_classorigin'],
                FilterQuery=options['filter_query'],
                FilterQueryLanguage=get_filterquerylanguage(options),
                MaxObjectCount=context.pull_max_cnt,
                PropertyList=resolve_propertylist(options['propertylist']))

        display_cim_objects(context, results, output_fmt,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)

    except ValueError as ve:
        raise click.ClickException('instance associators failed because '
                                   'FilterQuery not allowed with traditional '
                                   'Associators. --use-pull: '
                                   '{}. Exception: {}: {}'
                                   .format(context.use_pull,
                                           ve.__class__.__name__, ve))


def cmd_instance_count(context, classname, options):
    """
    Get the number of instances of each class in the namespace
    """
    output_fmt = validate_output_format(context.output_format, 'TABLE')

    # Differs from class find because it classname is optional.
    # If None, select all
    if classname is None:
        classname = '*'

    # Create list of namespaces from the option or from all namespaces
    ns_names = get_namespaces(context, options['namespace'])

    ns_cln_tuples = []  # a list of tuples of namespace, classname
    for namespace in ns_names:
        # Get all classes in Namespace
        try:
            # Set cmd options that are required for this command.
            # 1. Always use deep_inheritance
            # 2. Set namespace to each namespace in loop
            options['deep_inheritance'] = True
            options['namespace'] = namespace
            options['names_only'] = True

            classnames = enumerate_classes_filtered(context, None, options)
        except Error as er:
            raise_pywbem_error_exception(er)

        if classnames:
            classlist = filter_namelist(classname, classnames, ignore_case=True)
            cl_tup = [(namespace, cln) for cln in classlist]
            ns_cln_tuples.extend(cl_tup)

    # sort since normal output for this command is  namespace, classname
    # alphabetic order.
    ns_cln_tuples.sort(key=lambda tup: (tup[0], tup[1]))

    display_data = []
    for tup in ns_cln_tuples:
        ns = tup[0]
        cln = tup[1]
        # Try block allows issues where enumerate does not properly execute
        # The totals may be wrong but at least it gets what it can.
        # This accounts for issues with some servers where there
        # are providers that return errors from the enumerate.
        try:
            inst_names = context.conn.EnumerateInstanceNames(cln, namespace=ns)
        except CIMError as ce:
            warning_msg('Server Error {} with {}:{}. Continuing.'
                        .format(ce, ns, cln))

        # Sum the number of instances with the defined classname.
        # this counts only classes with that specific classname and not
        # subclasses
        clnl = cln.lower()
        count = sum(1 for inst_name in inst_names
                    if (inst_name.classname.lower() == clnl))

        if count != 0:
            display_tuple = (ns, cln, count)
            display_data.append(display_tuple)

    # If sort set, resort by count size
    if options['sort']:
        display_data.sort(key=lambda x: x[2])

    headers = ['Namespace', 'Class', 'count']
    rows = []
    if display_data:
        for item in display_data:
            rows.append([item[0], item[1], item[2]])

    context.spinner_stop()
    click.echo(format_table(rows, headers,
                            title='Count of instances per class',
                            table_format=output_fmt))


def cmd_instance_query(context, query, options):
    """Execute the query defined by the inputs"""
    output_fmt = validate_output_format(context.output_format, ['CIM', 'TABLE'])

    try:
        results = context.conn.PyWbemcliQueryInstances(
            options['query_language'],
            query,
            namespace=options['namespace'],
            MaxObjectCount=context.pull_max_cnt)

        display_cim_objects(context, results, output_fmt,
                            summary=options['summary'], sort=True)

    except Error as er:
        raise_pywbem_error_exception(er)


def cmd_instance_shrub(context, instancename, options):
    """
    Display the association information defined by instancename as a tree
    showing the various steps to get through the roles, references, etc. to
    return the names of associated instances.
    """
    try:
        instancepath = get_instancename(context, instancename, options)

        # Collect the data for the shrub
        shrub = AssociationShrub(context, instancepath,
                                 Role=options['role'],
                                 AssocClass=options['assoc_class'],
                                 ResultRole=options['result_role'],
                                 ResultClass=options['result_class'],
                                 verbose=context.verbose,
                                 fullpath=options['fullpath'])

        # display the shrub
        context.spinner_stop()
        shrub.display_shrub(context.output_format, options['summary'])
    except Error as er:
        raise_pywbem_error_exception(er)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility Functions applicable across multiple components of pywbemcli.
"""

from __future__ import print_function, absolute_import

import six
import click


__all__ = []


def _ensure_unicode(obj):
    """
    If the input object is a string, make sure it is returned as a
    :term:`unicode string`, as follows:

    * If the input object already is a :term:`unicode string`, it is returned
      unchanged.
    * If the input string is a :term:`byte string`, it is decoded using UTF-8.
    * Otherwise, the input object was not a string and is returned unchanged.

    Copied from pywbem
    """
    if isinstance(obj, six.binary_type):
        return obj.decode("utf-8")
    return obj


def _to_unicode(obj):
    """
    Convert the input binary string to a :term:`unicode string`.
    The input object must be a byte string.
    Use this if there is a previous test about the string type.

    Copies from pywbem
    """
    return obj.decode("utf-8")


def _eq_name(name1, name2):
    """
    Test two CIM names for equality.

    The comparison is performed case-insensitively.

    One or both of the names may be `None`; Two names that are `None` are
    considered equal.
    """
    if name1 is None:
        return name2 is None
    if name2 is None:
        return False
    return name1.lower() == name2.lower()


def deprecation_warning(msg, ctx_obj):
    """
    Display a deprecation warning on stderr, unless disabled via
    ctx_obj.deprecation_warnings.
    """
    if ctx_obj.deprecation_warnings:
        click.echo(msg, err=True)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pywbemcli supports a very limited number of configuration variables that
influence certain specific behavior.

These configuration variables are read by pywbem only after its modules have
been loaded, so they can be modified by the user directly after importing
pywbem. For example:

::

    import pywbemcli
    pywbem.DEFAULT_CONNECTION_TIMEOUT = 40

Note that the pywbem source file defining these variables should not be changed
by the user. Instead, the technique shown in the example above should be used to
modify the configuration variables.

Note: Due to limitations of the documentatin tooling, the following
configuration variables are shown in the ``pywbemcli.config`` namespace.
However, they should be used from the ``pywbemcli`` namespace.
"""

# This module is meant to be safe for 'import *'.

# Issue 224: Some strings are unicode because of this issue.


__all__ = ['DEFAULT_CONNECTION_TIMEOUT',
           'DEFAULT_NAMESPACE', 'PYWBEMCLI_PROMPT', 'PYWBEMCLI_HISTORY_FILE',
           'DEFAULT_MAXPULLCNT', 'MAX_TIMEOUT', 'DEFAULT_URL_SCHEME',
           'USE_TERMINAL_WIDTH', 'DEFAULT_TABLE_WIDTH']

#: Default value in seconds for a WBEMConnection to timeout if the value
#: is not set by an input parameter.
#: Positive integer.
DEFAULT_CONNECTION_TIMEOUT = 30

#: Specifies the default namespace uses if no default namespace is defined
#: on the cmd line, environment variable, or a config file.
DEFAULT_NAMESPACE = u'root/cimv2'

#: Specifies the default query language to be used for exedquery operations
#: when a query language is not specified in the request or config.
DEFAULT_QUERY_LANGUAGE = u'DMTF:CQL'

#: Characters for cmdline prompt when the pywbemcli repl is executing.
#: The prompt is presented at the beginning of a line awaiting a command
#: input.
#: The prompt MUST BE Unicode (prompt-toolkit requirement)
PYWBEMCLI_PROMPT = u'pywbemcli> '

#: File path of history file for interactive mode.
#: If the file name starts with tilde (which is handled by the shell, not by
#: the file system), it is properly expanded.
PYWBEMCLI_HISTORY_FILE = '~/.pywbemcli_history'

#: Default uri scheme if none is provided.  Thus if a server uri without
#: scheme component is provided, this is the default prepended to the
#: uri.

DEFAULT_URL_SCHEME = 'https'

#: Default pull MaxObjectCount if none is provided.  This is the maximum
#: number of objects per request that will be returned if the server
#: uses the pull operations for EnumerateInstances, AssociatorInstances,
#: etc. Set to the same default as used by pywbem.
DEFAULT_MAXPULLCNT = 1000

#: Maximum allowed connection timeout in seconds.  The environment will not
#: allow a connection timeout value larger than this on the command line or
#: internal option for timeout.
MAX_TIMEOUT = 300

#: If True, the table formatter uses the terminal width as the maximum width
#: of a table. If False it uses the DEFAULT_TABLE_WIDTH config variable. This
#: sets the maximum width of table output. The table formatter tries to
#: build any table output within this character width.
USE_TERMINAL_WIDTH = True

#: Default maximum character width of tables if USE_TERMINAL_WIDTH is NOT set.
#: If this variable is an integer, that is the maximum width. If None, tables
#: are output with no limit on width.
DEFAULT_TABLE_WIDTH = 150


#: If True, the auto-suggestion capability is enabled in the interactive
#: mode.  This capability uses the history file to provide suggestions for
#: the command file in addition to other auto-complete capabilities
#: If False, auto-suggestion is disabled.
USE_AUTOSUGGEST = True



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions to persist and restore the connections table.  This works from
a file that maintains defined connections. If any exist they are
loaded at startup and available through an interactive session.  Methods
are provided to create, delete, and view existing connections. A set method
allows setting the current active connection into the repository.
"""

from __future__ import absolute_import, print_function

import os
import yaml
import six
import yamlloader
import click

from ._pywbem_server import PywbemServer

if six.PY2:
    import codecs  # pylint: disable=wrong-import-order


# Normal definition for location of connection file
DEFAULT_CONNECTIONS_FILE = 'pywbemcli_connection_definitions.yaml'

LOCAL_CONNECTIONS_PATH = os.path.join(os.getcwd(), DEFAULT_CONNECTIONS_FILE)
HOME_CONNECTIONS_PATH = os.path.join(os.path.expanduser("~"),
                                     DEFAULT_CONNECTIONS_FILE)

BAK_FILE_SUFFIX = 'bak'


class ConnectionRepository(object):
    # pylint: disable=useless-object-inheritance
    """
    Manage the set of connections defined.  The data for the predefined
    connection exists on disk between pywbemcli sessions and within an
    instance of ConnectionRepository while pywbemcli is running.

    ConnectionRepository lazy load  gets the repository from the disk so that
    the repository is not in memory until one of the access methods is
    executed or a write is executed.  This also allows creation of new
    connections files using the connection save command.
    """
    # Class variables
    # The YAML section names in the connection yaml file
    # Name of the YAML group that contains named connection definitions.
    connections_group_name = 'connection_definitions'
    # Name of the YAML group that contains the default connection name
    default_connection_grp_name = 'default_connection_name'

    def __init__(self, connections_file):
        """
        Initialize the object instance if it has not already been initialized
        (class level variable is not None)by reading the connection file.

        The __init__ does not load the connections file. That is done for each
        method that accesses the file to read or write.

        Parameters:

          connections_file (:term:`string`):
            File path defining the location of the connections file.  This
            is optional and must define the location of an existing file.
        """

        # Dictionary of named connections if the connection file is loaded
        # where each value is an instance of the class PywbemServer
        self._pywbemcli_servers = {}

        # Flag indicating whether the connections file has been loaded
        self._loaded = False
        self._connections_file = connections_file

        # default connection name Must be the name of a
        # connection in the connections file or None.
        self._default_connection_name = None

    @property
    def connections_file(self):
        """
        Get the current connections file

        Returns:
            :term:`string` containing the file path of the connections file
        """
        return self._connections_file

    @property
    def default_connection_name(self):
        """
        Return the name of connection defined in the connections file that
        is the default connection, the connection that is selected on
        pywbemcli startup if no connection name or other connection definition
        is specified.

        If None, there is no current default connection name

        Returns:
            :term:`string` containing the name of the default connection
            defined in the connections file.
        """
        return self._default_connection_name

    def file_exists(self):
        """
        Test if the connection file exists

        Returns:
            (:class:`py:bool`) True if the connections_file exists and False if
            it does not exist
        """
        return os.path.isfile(self.connections_file)

    def __str__(self):
        """
        Return  a string containing connections file name and count of items in
        connections file
        """
        if self.file_exists():
            status = 'exists'
            self._load_connections_file()
            length = len(self)
        else:
            status = "does not exist"
            length = 0

        return('connections_file: {0} {1}. {2} servers defined.'.format(
            self.connections_file, status, length))

    def __repr__(self):
        """
        Return a string representation of the
        servers dictionary that is suitable for debugging and other key
        parameters. Always tries to load connection file before displaying.
        """
        self._load_connections_file()
        # Fails if connection file does not exist.
        items = ["{0}: {1}".format(k, v)
                 for k, v in six.iteritems(self._pywbemcli_servers)]
        items_str = ', '.join(items)
        return "{0.__class__.__name__}({{{1}}}, connections_file {2})]". \
            format(self, items_str, self._connections_file)

    def __contains__(self, key):
        """Load the repository if necessary and return True if key found"""
        self._load_connections_file()
        return key in self._pywbemcli_servers

    def __getitem__(self, key):
        self._load_connections_file()
        return self._pywbemcli_servers[key]

    def __len__(self):
        self._load_connections_file()
        return len(self._pywbemcli_servers)

    def __iter__(self):
        self._load_connections_file()
        return six.iterkeys(self._pywbemcli_servers)

    def items(self):
        """
        Return a list of the items in the server repo
        """
        self._load_connections_file()
        return list(self.__iteritems__())

    def keys(self):
        """
        Return a list of the dictionary keys.
        """
        self._load_connections_file()
        return list(self.iterkeys())

    def __iteritems__(self):  # pylint: disable=no-self-use
        """"""
        self._load_connections_file()
        return six.iteritems(self._pywbemcli_servers)

    def iterkeys(self):
        """
        Return an iterator through the dictionary keys in their original
        case, preserving the original order of items.
        """
        self._load_connections_file()
        for item in six.iterkeys(self._pywbemcli_servers):
            yield item

    def _load_connections_file(self):
        """
        If there is a file, read it in and install into the dictionary.
        This file is read only once.

        Raises:
          IOError: repository does not exist

          ValueError: Error in reading repository
        """
        if self._loaded:
            return

        if os.path.isfile(self._connections_file):
            with self._open_file(self._connections_file, 'r') as _fp:
                try:
                    dict_ = yaml.safe_load(_fp)
                    # put all the connection definitions into a group
                    # in the connection file
                    connections_dict = dict_[
                        ConnectionRepository.connections_group_name]

                    self._default_connection_name = dict_[
                        ConnectionRepository.default_connection_grp_name]
                    try:
                        for name, svr in six.iteritems(connections_dict):
                            self._pywbemcli_servers[name] = \
                                PywbemServer.create(
                                    replace_underscores=True, **svr)
                            self._loaded = True
                    except KeyError as ke:
                        raise KeyError("Items missing from record {0} in "
                                       "connections file. Exception {1}".format
                                       (ke, self._connections_file))
                    except TypeError as te:
                        raise TypeError('Invalid object type in connections '
                                        'file: "{0}"; server name: "{1}". '
                                        'Item: {2}'.
                                        format(self._connections_file, name,
                                               te))
                except ValueError as ve:
                    raise ValueError("Invalid YAML in connections file {0}. "
                                     "Exception {1}".format
                                     (self._connections_file, ve))
        # Apply IO error for non-existent file
        # TODO. IOError can also mean disk full in general
        else:
            raise IOError("Connections file {} does not exist.".
                          format(self.connections_file))

    def add(self, name, svr_definition):
        """
        Add a new connection to the connections repository or replace an
        existing connection.  Users of this method should check before add if
        they do not want to replace an existing entry.

        Parameter:

          svr_definition (:class:`ServerDefinition`):
            An instance of ServerDefinition that contains the data that will be
            added to the connections file.

        Raises:
            click.Abort if the load of the connections file returns an
            exception
        """

        # If the file does exists get it. Otherwise the write will create it
        # TODO: we process errors here differently than in read mode because
        # de use click and abort. Should we do this in common.
        # Also, why not make abort_click a common function with message.
        if os.path.isfile(self._connections_file):
            try:
                self._load_connections_file()
            # All other errors cases abort.
            except Exception as e:
                click.echo("Fatal error loading YAML file {0}. "
                           "Exception: {1}: {2}".
                           format(self.connections_file, e.__class__.__name__,
                                  e),
                           err=True)
                raise click.Abort()

        self._pywbemcli_servers[name] = svr_definition
        self._write_file()

    def delete(self, name):
        """
        Delete a definition from the connections repository.

        Parameters:

          name (:term:`string):
            Name of the connection to delete

        Raises:
            KeyError: No server definition exists in the connections file
            with this name

            ValueError: Connection file is not valid YAML

            IOError: The connections file does not exist.

        """
        self._load_connections_file()
        del self._pywbemcli_servers[name]

        # Remove default_name if it is the name being deleted
        if name == self.default_connection_name:
            self._default_connection_name = None
        self._write_file()

    @staticmethod
    def _open_file(filename, file_mode='w'):
        """
        A static convenience function that performs the open of the connection
        definitions file correctly for different versions of Python.

        This covers the issue where the file should be opened in text mode but
        that is done differently in Python 2 and Python 3.

        The returned file-like object must be closed by the caller.

        Parameters:

          filename (:term:`string`):
            Name of the file where the recorder output will be written

          file_mode (:term:`string`):
            Optional file mode.  The default is 'w' which overwrites any
            existing file.  if 'a' is used, the data is appended to any
            existing file.

        Returns:

          File-like object.
        """
        if six.PY2:
            # Open with codecs to define text mode
            return codecs.open(filename, mode=file_mode, encoding='utf-8')

        return open(filename, file_mode, encoding='utf8')

    def _write_file(self):  # pylint: disable=no-self-use
        """
        Write the connections file if one has been loaded.
        If the dictionary is empty, it deletes the file.

        If there is an existing file it is moved to filename.bak and a new
        current file written.
        """
        conn_dict = {}
        if self._pywbemcli_servers:
            if self._pywbemcli_servers:
                conn_dict = \
                    {name: value.to_dict() for name, value in
                     self._pywbemcli_servers.items()}

            # build dictionary for yaml output
            yaml_dict = {ConnectionRepository.connections_group_name: conn_dict,
                         ConnectionRepository.default_connection_grp_name:
                             self.default_connection_name}

            # Write to tmpfile and if successful create backup file and
            # move the tmpfile to be the new connections file contents.
            tmpfile = '{}.tmp'.format(self._connections_file)

            with self._open_file(tmpfile, 'w') as _fp:
                data = yaml.dump(yaml_dict,
                                 encoding=None,
                                 allow_unicode=True,
                                 default_flow_style=False,
                                 indent=4,
                                 Dumper=yamlloader.ordereddict.CSafeDumper)
                data = data.replace('\n\n', '\n')  # YAML dump dups newlines
                _fp.write(data)
                _fp.flush()

        # create bak file and then rename tmp file
        if os.path.isfile(self._connections_file):
            bakfile = '{0}.{1}'.format(self._connections_file, BAK_FILE_SUFFIX)
            if os.path.isfile(bakfile):
                os.remove(bakfile)
            if os.path.isfile(self._connections_file):
                os.rename(self._connections_file, bakfile)

        if self._pywbemcli_servers:
            os.rename(tmpfile, self._connections_file)

    def set_default_connection(self, connection_name):
        """
        Set the connection defined by connection_name to be the current
        connection in the connections file.

        This is accomplished by modifying the "current_connection" entry
        and rewriting the file.

        Parameters:

          connection_name (:term:`string`):
            The name of an existing connection in the connection file.

        Raises:
          ValueError: if connection_name does not exist in the connection file.
        """
        self._load_connections_file()
        if connection_name in self._pywbemcli_servers:
            self._default_connection_name = connection_name
            self._write_file()

        else:
            raise ValueError('Connection name "{}" does not exist in '
                             'connection repository {}'
                             .format(connection_name, self.connections_file))

    def get_default_connection_name(self):
        """
        Returns the name of the current connection in the connections file.
        This may be the name of a connection in the connections file or
        None if no connection is defined as the current connection.

        Returns:
          :term:`string`: Name of the default connection if one
          exists in the file; otherwise None

        Raises:
            ValueError: Connection file is not valid YAML

            IOError: The connections file does not exist.

        """
        if self.file_exists():
            self._load_connections_file()
            return self.default_connection_name
        return None



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common Functions applicable across multiple components of pywbemcli
"""

from __future__ import absolute_import, print_function, unicode_literals

import re
from collections import OrderedDict
import six
import click

import pywbem
from pywbem import WBEMServer, configure_loggers_from_string

from .config import DEFAULT_URL_SCHEME, DEFAULT_CONNECTION_TIMEOUT, \
    DEFAULT_NAMESPACE, MAX_TIMEOUT
from ._pywbemcli_operations import PYWBEMCLIConnection, PYWBEMCLIFakedConnection

WBEM_SERVER_OBJ = None

PYWBEMCLI_LOG = 'pywbemcli.log'


def _raise_typeerror(name, value, rqd_type):
    """
    Generate a TypeError for a property in pywbem_server setter that has an
    invalid type
    """
    raise TypeError('Property "{0}" value: {1} must be type: "{2}", not type: '
                    '"{3}"'.format(name, value, rqd_type, type(value)))


def _validate_server_url(server):
    """
    Validate  and possibly complete the wbemserver url provided.

    Parameters:

      server (string):
        url of the WBEMServer to which connection is being made including
        scheme, hostname/IPAddress, and optional port

    Returns:
        The input url or url extended to use DEFAULT_SCHEME as the
        scheme if not is provided

    Raises:
        click.CLICKException if scheme invalid
    """
    if server[0] == '/':
        url = server

    elif re.match(r"^https{0,1}://", server) is not None:
        url = server

    # TODO Future: We are trying to make exceptions in these support classes
    # independent of click so this should become ValueError
    elif re.match(r"^[a-zA-Z0-9]+://", server) is not None:
        raise click.ClickException('Invalid scheme on server argument. {}'
                                   ' Use "http" or "https"'.format(server))
    else:
        url = "{scheme}://{host}".format(
            scheme=DEFAULT_URL_SCHEME,
            host=DEFAULT_URL_SCHEME)

    return url


class PywbemServer(object):
    # pylint: disable=too-many-instance-attributes, useless-object-inheritance
    """
    Envelope for connections with WBEM Server incorporates both the
    pywbem WBEMConnection and pywbem WBEMServer classes.

    The constructor is separated from the connection method to allow the
    object to be contstructed early but actually create the connection and
    WBEMServer objects only when required.  This allows parameters such
    as the password to be requested only when a connection is made.

    This class also holds the variables that determine whether the connection
    will use the pull operations or traditional operations
    """
    # ISSUE: #658 - Reorganize the location for these variable names separate
    # from the PywbemServer class.
    # The following class level variables are the names for the env variables
    # where server connection information are be saved and used as alternate
    # input sources for pywbemcli arguments and options.
    server_envvar = 'PYWBEMCLI_SERVER'
    name_envvar = 'PYWBEMCLI_NAME'
    user_envvar = 'PYWBEMCLI_USER'
    password_envvar = 'PYWBEMCLI_PASSWORD'
    defaultnamespace_envvar = 'PYWBEMCLI_DEFAULT_NAMESPACE'
    timeout_envvar = 'PYWBEMCLI_TIMEOUT'
    keyfile_envvar = 'PYWBEMCLI_KEYFILE'
    certfile_envvar = 'PYWBEMCLI_CERTFILE'
    verify_envvar = 'PYWBEMCLI_VERIFY'
    ca_certs_envvar = 'PYWBEMCLI_CA_CERTS'
    timestats_envvar = 'PYWBEMCLI_TIMESTATS'
    use_pull_envvar = 'PYWBEMCLI_USE_PULL'
    pull_max_cnt_envvar = 'PYWBEMCLI_PULL_MAX_CNT'
    mock_server_envvar = 'PYWBEMCLI_MOCK_SERVER'
    log_envvar = 'PYWBEMCLI_LOG'
    # The following exports are not part of the pywbem_server container
    pdb_envvar = 'PYWBEMCLI_PDB'
    deprecation_warnings_envvar = 'PYWBEMCLI_DEPRECATION_WARNINGS'
    termwidth_envvar = 'PYWBEMCLI_TERMWIDTH'
    connections_file_envvar = 'PYWBEMCLI_CONNECTIONS_FILE'

    def __init__(self, server=None, default_namespace=DEFAULT_NAMESPACE,
                 name='default', user=None, password=None,
                 timeout=DEFAULT_CONNECTION_TIMEOUT, verify=None, use_pull=None,
                 certfile=None, keyfile=None, ca_certs=None, mock_server=None):
        """
        Create  a PywbemServer object. This contains the configuration
        and operation information to create a connection to the server
        and execute cim_operations on the server.
        """

        if server and mock_server:
            raise ValueError('Simultaneous "--server" and '
                             '"--mock-server" not allowed. Server: {}, '
                             'mock_server {}'.format(server, mock_server))
        self.server = server
        self.mock_server = mock_server
        self.name = name
        self.default_namespace = default_namespace
        self.user = user
        self.password = password
        self.timeout = timeout
        self.use_pull = use_pull
        self.verify = verify
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_certs = ca_certs

        # dynamically created in create_connection
        self._wbem_server = None
        self._conn = None

    def __str__(self):
        return 'PywbemServer(url={} name={})'.format(self.server, self.name)

    def __repr__(self):
        return 'PywbemServer(server={} name={} ns={} user={} ' \
               'password={} timeout={} use_pull={} verify={} certfile={} ' \
               'keyfile={} ca_certs={}  ' \
               'mock_server={!r} wbem_server {!r})' \
               .format(self.server, self.name, self.default_namespace,
                       self.user, self.password, self.timeout, self.use_pull,
                       self.verify, self.certfile, self.keyfile, self.ca_certs,
                       self.mock_server, self.wbem_server)

    @property
    def server(self):
        """
        :term:`string`: Scheme with Hostname or IP address of the WBEM Server.
        """
        return self._server

    @server.setter
    def server(self, server):
        """Setter method; for a description see the getter method."""

        # pylint: disable=attribute-defined-outside-init
        if server:
            self._server = _validate_server_url(server)
        else:
            self._server = server

    @property
    def mock_server(self):
        """
        list of :term:`string`: list of file paths for mock server setup.
        """
        return self._mock_server

    @mock_server.setter
    def mock_server(self, mock_server):
        """Setter method; for a description see the getter method."""
        if mock_server:
            if not isinstance(mock_server, (list, six.string_types)):
                _raise_typeerror("mock_server", mock_server, 'list, string')
            for ms in mock_server:
                if not isinstance(ms, six.string_types):
                    _raise_typeerror("mock_server item", mock_server, 'string')
        # assure this is list type in yaml output
        if mock_server is None:
            mock_server = []
        # pylint: disable=attribute-defined-outside-init
        self._mock_server = mock_server

    @property
    def name(self):
        """
        :term:`string`: Defines a name for this connection object.
        """
        return self._name

    @name.setter
    def name(self, name):
        """Setter method; for a description see the getter method."""
        if name:
            if not isinstance(name, six.string_types):
                _raise_typeerror("name", name, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._name = name

    @property
    def user(self):
        """
        :term:`string`: user name on the WBEM Server.
        """
        return self._user

    @user.setter
    def user(self, user):
        """Setter method; for a description see the getter method."""

        if user and not isinstance(user, six.string_types):
            _raise_typeerror("user", user, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._user = user

    @property
    def password(self):
        """
        :term:`string`: Password for this user on this WBEM Server.
        """
        return self._password

    @password.setter
    def password(self, password):
        """Setter method; for a description see the getter method."""
        if password and not isinstance(password, six.string_types):
            _raise_typeerror("password", password, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._password = password

    @property
    def default_namespace(self):
        """
        :term:`string`: Namespace to be used as default  for requests.
        """
        return self._default_namespace

    @default_namespace.setter
    def default_namespace(self, default_namespace):
        """Setter method; for a description see the getter method."""
        if default_namespace and not isinstance(default_namespace,
                                                six.string_types):
            _raise_typeerror("default-namespace", default_namespace, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._default_namespace = default_namespace

    @property
    def timeout(self):
        """
        :term:`int`: Connection timeout to be used on requests in seconds
        """
        if self._timeout:
            assert isinstance(self._timeout, int)
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Setter method; for a description see the getter method."""
        if timeout is None:   # disallow None
            raise ValueError('timeout option of None not allowed')
        if not isinstance(timeout, int):
            _raise_typeerror("timeout", timeout, 'integer')
        if not 0 < timeout <= MAX_TIMEOUT:
            raise ValueError('Timeout option "{0}" out of range {1} to {2} sec'
                             .format(timeout, 0, MAX_TIMEOUT))
        # pylint: disable=attribute-defined-outside-init
        self._timeout = timeout

    @property
    def use_pull(self):
        """
        :term:`string`: Connection timeout to be used on requests in seconds
        """
        return self._use_pull

    @use_pull.setter
    def use_pull(self, use_pull):
        """Setter method; for a description see the getter method."""

        if use_pull is None or isinstance(use_pull, bool):
            # pylint: disable=attribute-defined-outside-init
            self._use_pull = use_pull
        else:
            _raise_typeerror("use-pull", use_pull, 'boolean')

    @property
    def verify(self):
        """
        :class:`py:bool`: Connection server verfication flag. If True
        server cert verified during connection.
        """
        return self._verify

    @verify.setter
    def verify(self, verify):
        """Setter method; for a description see the getter method."""
        if verify and not isinstance(verify, bool):
            _raise_typeerror("verify", verify, 'boolean')
        # pylint: disable=attribute-defined-outside-init
        self._verify = verify

    @property
    def certfile(self):
        """
        :term:`string`: certificate for server or None if parameter not
        provided on input
        """
        if self._certfile:
            assert isinstance(self._certfile, six.string_types)
        return self._certfile

    @certfile.setter
    def certfile(self, certfile):
        """Setter method; for a description see the getter method."""
        if certfile and not isinstance(certfile, six.string_types):
            _raise_typeerror("certfile", certfile, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._certfile = certfile

    @property
    def keyfile(self):
        """
        :term:`string`: keyfile or None if no keyfile parameter input
        """
        if self._keyfile:
            assert isinstance(self._keyfile, six.string_types)
        return self._keyfile

    @keyfile.setter
    def keyfile(self, keyfile):
        """Setter method; for a description see the getter method."""
        if keyfile and not isinstance(keyfile, six.string_types):
            _raise_typeerror("keyfile", keyfile, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._keyfile = keyfile

    @property
    def ca_certs(self):
        """
        :term:`string`: String that defines certs for server validation"
        """
        if self._ca_certs:
            assert isinstance(self._ca_certs, six.string_types)
        return self._ca_certs

    @ca_certs.setter
    def ca_certs(self, ca_certs):
        """Setter method; for a description see the getter method."""
        if ca_certs and not isinstance(ca_certs, six.string_types):
            _raise_typeerror("ca_certs", ca_certs, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._ca_certs = ca_certs

    @property
    def conn(self):
        """
        :class:`~pywbem.WBEMConnection` WBEMConnection to be used for requests.
        """
        # This is created in wbemserver and retained there.
        return self.wbem_server.conn if self.wbem_server else None

    @property
    def wbem_server(self):
        """
        :class:`~pywbem.WBEMConnection` WBEMServer instance to be used for
        requests.
        """
        return self._wbem_server

    def password_prompt(self, ctx):
        """
        Request password from console.
        """
        if self.user:
            ctx.spinner_stop()
            password = click.prompt(
                "Enter password (user {user})" .format(user=self.user),
                hide_input=True,
                confirmation_prompt=False, type=str, err=True)
            ctx.spinner_start()
            # pylint: disable=attribute-defined-outside-init
            self.password = password
        else:
            # TODO: Again we want to isolate the click excpetions in future
            raise click.ClickException("{cmd} requires user/password, but "
                                       "no password provided."
                                       .format(cmd=ctx.invoked_subcommand))

    def get_password(self, ctx):
        """
        Conditional password prompt function  Prompts for password only if
        there  is a defined user and no password.
        """
        if self.user and not self.password:
            self.password_prompt(ctx)

    def to_dict(self):
        """
        Create dictionary from instance for persisting the connection. All
        key names that include _ are modified to - so that the dictionary
        reflects the general option names.
        """

        return OrderedDict({"name": self.name,
                            "server": self.server,
                            "user": self.user,
                            "password": self.password,
                            "default-namespace": self.default_namespace,
                            "timeout": self.timeout,
                            "use_pull": self.use_pull,
                            "verify": self.verify,
                            "certfile": self.certfile,
                            "keyfile": self.keyfile,
                            "ca-certs": self.ca_certs,
                            "mock-server": self.mock_server})

    @staticmethod
    def create(replace_underscores=False, **kwargs):
        """Create PywbemServer object from kwargs. If replace_underscore is
        True, replace any -  in names with _
        """
        kwargsout = {}
        if replace_underscores:
            kwargsout = {k.replace('-', '_'): v for k, v in kwargs.items()}
        return PywbemServer(**kwargsout)

    def reset(self):
        """ Reset the connection attributes of this pywbem server so that the
        connection must be restablished
        """
        self._conn = None
        self._wbem_server = None

    def create_connection(self, log=None, use_pull=None,
                          verbose=None, timestats=None):
        """
        Initiate a WBEB connection, via PyWBEM api. Arguments for
        the request are the parameters required by the pywbem
        WBEMConnection constructor that are not available within this
        class.

        If self.mock_server is set, a mock connection is created instead
        of a genuine connection to a server.
        See the pywbem.WBEMConnection class for more details on the parameters.

        Return:
            pywbem.WBEMConnection: Connection object that can be used to execute
            other pywbem cim_operation requests

        Raises:
           ValueError: if server paramer is invalid or other issues with
           the input values
        """
        if self._mock_server:
            if self.conn is None:
                conn = PYWBEMCLIFakedConnection(
                    default_namespace=self.default_namespace,
                    use_pull_operations=use_pull,
                    stats_enabled=timestats)
                try:
                    self._wbem_server = WBEMServer(conn)
                    conn.build_repository(conn,
                                          self._wbem_server,
                                          self._mock_server,
                                          verbose)
                except click.Abort:
                    raise
                except Exception as exc:  # pylint: disable=broad-except
                    click.echo(
                        "Building the mock repository failed: {}".format(exc),
                        err=True)

        else:  # mock_server does not exist
            if not self.server:
                raise click.ClickException('No server found. Cannot '
                                           'connect.')
            if self.keyfile is not None and self.certfile is None:
                ValueError('keyfile option requires certfile option')

            creds = (self.user, self.password) if self.user or \
                self.password else None

            # If client cert and key provided, create dictionary for
            # wbem connection certs (WBEMConnection takes dictionary for this
            # info)
            x509_dict = None
            if self.certfile is not None:
                x509_dict = {"cert_file": self.certfile}
                if self.keyfile is not None:
                    x509_dict.update({'key_file': self.keyfile})

            # Create the WBEMConnection object and the _wbem_server object

            # Negate verify to no_verification
            if self.verify is None:
                no_verification = self.verify
            else:
                no_verification = not self.verify

            # Convert ca_certs command line option to ca_certs parameter
            if getattr(pywbem, 'PYWBEM_USES_REQUESTS', False):
                if self.ca_certs == 'certifi':
                    ca_certs = None
                else:
                    ca_certs = self.ca_certs
            else:
                ca_certs = self.ca_certs

            try:
                conn = PYWBEMCLIConnection(
                    self.server, creds,
                    default_namespace=self.default_namespace,
                    no_verification=no_verification,
                    x509=x509_dict, ca_certs=ca_certs,
                    timeout=self.timeout,
                    use_pull_operations=use_pull,
                    stats_enabled=timestats)
            except IOError as exc:
                raise click.ClickException(
                    'Cannot create connection to {}: {}'.
                    format(self.server, exc))

            # Create a WBEMServer object
            self._wbem_server = WBEMServer(conn)

        if log:
            try:
                configure_loggers_from_string(log,
                                              log_filename=PYWBEMCLI_LOG,
                                              connection=conn, propagate=True)
            except ValueError as ve:
                raise click.ClickException('Logger configuration error. input: '
                                           '{}. Exception: {}'.format(log, ve))



#!/usr/bin/env python

# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Setup script for pywbemtools project.
"""

import os
import re
import setuptools


def get_version(version_file):
    """
    Execute the specified version file and return the value of the __version__
    global variable that is set in the version file.

    Note: Make sure the version file does not depend on any packages in the
    requirements list of this package (otherwise it cannot be executed in
    a fresh Python environment).
    """
    with open(version_file, 'r') as fp:
        version_source = fp.read()
    _globals = {}
    exec(version_source, _globals)  # pylint: disable=exec-used
    return _globals['__version__']


def get_requirements(requirements_file):
    """
    Parse the specified requirements file and return a list of its non-empty,
    non-comment lines. The returned lines are without any trailing newline
    characters.
    """
    with open(requirements_file, 'r') as fp:
        lines = fp.readlines()
    reqs = []
    for line in lines:
        line = line.strip('\n')
        if not line.startswith('#') and line != '':
            reqs.append(line)
    return reqs


def read_file(a_file):
    """
    Read the specified file and return its content as one string.
    """
    with open(a_file, 'r') as fp:
        content = fp.read()
    return content


# pylint: disable=invalid-name
requirements = get_requirements('requirements.txt')
install_requires = [req for req in requirements
                    if req and not re.match(r'[^:]+://', req)]
dependency_links = [req for req in requirements
                    if req and re.match(r'[^:]+://', req)]
package_version = get_version(os.path.join('pywbemtools', '_version.py'))

# Docs on setup():
# * https://docs.python.org/2.7/distutils/apiref.html?
#   highlight=setup#distutils.core.setup
# * https://setuptools.readthedocs.io/en/latest/setuptools.html#
#   new-and-changed-setup-keywords
setuptools.setup(
    name='pywbemtools',
    version=package_version,
    packages=[
        'pywbemtools',
    ],
    entry_points={
        'console_scripts': [
            'pywbemcli = pywbemtools.pywbemcli.pywbemcli:cli',
        ],
    },
    include_package_data=True,  # as specified in MANIFEST.in
    install_requires=install_requires,
    dependency_links=dependency_links,

    description='Python client tools to work with WBEM Servers using the '
    'PyWBEM API.',
    long_description=read_file('README_PYPI.rst'),
    long_description_content_type='text/x-rst',
    license='Apache License, Version 2.0',
    author='Karl Schopmeyer, Andreas Maier',
    author_email='k.schopmeyer@swbell.net, maiera@de.ibm.com',
    maintainer='Karl Schopmeyer, Andreas Maier',
    maintainer_email='k.schopmeyer@swbell.net, maiera@de.ibm.com',
    url='https://github.com/pywbem/pywbemtools',
    project_urls={
        'Bug Tracker': 'https://github.com/pywbem/pywbemtools/issues',
        'Documentation': 'https://pywbemtools.readthedocs.io/en/latest/',
        'Source Code': 'https://github.com/pywbem/pywbemtools',
    },

    options={'bdist_wheel': {'universal': True}},
    zip_safe=True,  # This package can safely be installed from a zip file
    platforms='any',

    # Keep these Python versions in sync with pywbemtools/__init__.py
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)
